"""Contains tests for the thermostat parameter descriptors."""

import pytest

from pyplumio.const import ATTR_INDEX, ATTR_OFFSET, ATTR_SIZE, ATTR_VALUE
from pyplumio.devices.ecomax import EcoMAX
from pyplumio.devices.thermostat import Thermostat
from pyplumio.frames.requests import SetThermostatParameterRequest
from pyplumio.parameters import ParameterValues
from pyplumio.parameters.thermostat import (
    ThermostatNumber,
    ThermostatNumberDescription,
    ThermostatSwitch,
    ThermostatSwitchDescription,
    get_thermostat_parameter_types,
)


@pytest.fixture(name="thermostat")
def fixture_thermostat(ecomax: EcoMAX) -> Thermostat:
    """Return an thermostat object."""
    return Thermostat(ecomax.queue, parent=ecomax)


async def test_mixer_parameter_create_request(thermostat: Thermostat) -> None:
    """Test create_request method for mixer parameter."""
    # Test with number.
    thermostat_number = ThermostatNumber(
        device=thermostat,
        description=ThermostatNumberDescription(name="test_number"),
        values=ParameterValues(value=10, min_value=0, max_value=20),
    )
    thermostat_number_request = await thermostat_number.create_request()
    assert isinstance(thermostat_number_request, SetThermostatParameterRequest)
    assert thermostat_number_request.data == {
        ATTR_INDEX: 1,
        ATTR_VALUE: 10,
        ATTR_OFFSET: 0,
        ATTR_SIZE: 1,
    }

    # Test with switch.
    thermostat_switch = ThermostatSwitch(
        device=thermostat,
        description=ThermostatSwitchDescription(name="test_switch"),
        values=ParameterValues(value=0, min_value=0, max_value=1),
    )
    thermostat_switch_request = await thermostat_switch.create_request()
    assert isinstance(thermostat_switch_request, SetThermostatParameterRequest)
    assert thermostat_switch_request.data == {
        ATTR_INDEX: 1,
        ATTR_VALUE: 0,
        ATTR_OFFSET: 0,
        ATTR_SIZE: 1,
    }


def test_get_thermostat_parameter_types() -> None:
    """Test ecoMAX parameter types getter."""
    parameter_types_all = get_thermostat_parameter_types()
    assert len(parameter_types_all) == 15



"""Contains tests for the connection classes."""

from __future__ import annotations

from asyncio import StreamReader, StreamWriter
from importlib import reload
import logging
import sys
from typing import Any, Final
from unittest.mock import AsyncMock, Mock, patch

import pytest
from serial import EIGHTBITS, PARITY_NONE, STOPBITS_ONE

import pyplumio.connection
from pyplumio.connection import Connection, SerialConnection, TcpConnection
from pyplumio.exceptions import ConnectionFailedError
from pyplumio.protocol import Protocol


@pytest.fixture(name="use_fast", params=(True, False))
def fixture_use_fast(request, monkeypatch, caplog) -> Any:
    """Try with and without serial-asyncio-fast package."""
    if not request.param:
        monkeypatch.setitem(sys.modules, "serial_asyncio_fast", None)

    with caplog.at_level(logging.INFO):
        reload(pyplumio.connection)

    message = "Using pyserial-asyncio-fast in place of pyserial-asyncio"
    if request.param:
        assert message in caplog.text
    else:
        assert message not in caplog.text

    return request.param


@pytest.fixture(name="stream_writer")
def fixture_stream_writer():
    """Return a mock of asyncio stream writer."""
    with patch("asyncio.StreamWriter", autospec=True) as mock_stream_writer:
        yield mock_stream_writer


@pytest.fixture(name="stream_reader")
def fixture_stream_reader():
    """Return a mock of asyncio stream reader."""
    with patch("asyncio.StreamReader", autospec=True) as mock_stream_reader:
        yield mock_stream_reader


@pytest.fixture(name="asyncio_open_connection")
def fixture_asyncio_open_connection(
    stream_reader: StreamReader, stream_writer: StreamWriter
):
    """Bypass opening asyncio connection."""
    with patch(
        "asyncio.open_connection", return_value=(stream_reader, stream_writer)
    ) as mock_connection:
        yield mock_connection


@pytest.fixture(name="serial_asyncio_open_serial_connection")
def fixture_serial_asyncio_open_serial_connection(
    use_fast, stream_reader: StreamReader, stream_writer: StreamWriter
):
    """Bypass opening serial_asyncio connection."""
    module = "serial_asyncio_fast" if use_fast else "serial_asyncio"
    with patch(
        f"{module}.open_serial_connection",
        return_value=(stream_reader, stream_writer),
    ) as mock_connection:
        yield mock_connection


@pytest.fixture(name="mock_protocol")
def fixture_mock_protocol():
    """Return a mock protocol object."""
    return AsyncMock(spec=Protocol, autospec=True)


@pytest.fixture(name="tcp_connection")
def fixture_tcp_connection(mock_protocol) -> TcpConnection:
    """Return at TCP connection object."""
    return TcpConnection(host=HOST, port=PORT, timeout=10, protocol=mock_protocol)


@pytest.fixture(name="serial_connection")
def fixture_serial_connection(mock_protocol) -> SerialConnection:
    """Return a serial connection object."""
    return SerialConnection(device="/dev/ttyUSB0", timeout=10, protocol=mock_protocol)


class DummyConnection(Connection):
    """Represents a dummy connection for tests."""

    async def _open_connection(self):
        """Open the connection."""


class TestConnection:
    """Contains test for Connection class."""

    @patch.object(
        DummyConnection, "_open_connection", return_value=("reader", "writer")
    )
    async def test_connect(self, mock_open_connection, mock_protocol) -> None:
        """Test a connection."""
        connection = DummyConnection(protocol=mock_protocol)
        assert connection.protocol is mock_protocol
        await connection.connect()
        mock_open_connection.assert_awaited_once()
        await connection.close()
        mock_protocol.shutdown.assert_called_once()

    @pytest.mark.parametrize(
        ("reconnect_on_failure", "expected_log"),
        [
            (True, "Can't connect to the device"),
            (False, None),
        ],
    )
    @patch.object(DummyConnection, "_open_connection", side_effect=(OSError, None))
    async def test_reconnect(
        self,
        mock_open_connection,
        caplog,
        reconnect_on_failure: bool,
        expected_log: str | None,
    ) -> None:
        """Test reconnect logic."""
        connection = DummyConnection(reconnect_on_failure=reconnect_on_failure)
        if reconnect_on_failure:
            with caplog.at_level(logging.ERROR):
                await connection.connect()

            assert expected_log in caplog.text
        else:
            with pytest.raises(ConnectionFailedError):
                await connection.connect()

        mock_open_connection.assert_awaited_once()

    @patch.object(DummyConnection, "close")
    @patch.object(DummyConnection, "connect")
    async def test_context_manager(self, mock_connect, mock_close) -> None:
        """Test context manager."""
        async with DummyConnection():
            ...

        mock_connect.assert_called_once()
        mock_close.assert_called_once()

    @patch.object(DummyConnection, "_connect")
    async def test_connection_lost(self, mock_connect, mock_protocol) -> None:
        """Test that connection lost callback calls reconnect."""
        connection = DummyConnection(protocol=mock_protocol)
        await connection.connect()
        on_connection_lost = mock_protocol.on_connection_lost.add.call_args.args[0]
        await on_connection_lost()
        assert mock_connect.call_count == 2

    async def test_getattr(self, mock_protocol) -> None:
        """Test that getattr is getting proxied to the protocol."""
        connection = DummyConnection(protocol=mock_protocol)
        mock_protocol.some_attr = "value"
        assert connection.some_attr == "value"
        mock_protocol.some_method = Mock(return_value="called")
        assert connection.some_method() == "called"


HOST: Final = "localhost"
PORT: Final = 8899


class TestTcpConnection:
    """Contains tests for TcpConnection class."""

    async def test_connection(
        self, tcp_connection: TcpConnection, asyncio_open_connection
    ) -> None:
        """Test TCP connection."""
        await tcp_connection.connect()
        asyncio_open_connection.assert_awaited_once_with(
            host=HOST, port=PORT, timeout=10
        )

    async def test_repr(self, tcp_connection: TcpConnection) -> None:
        """Test serializable representation."""
        assert (
            repr(tcp_connection)
            == f"TcpConnection(host={HOST}, port={PORT}, kwargs={{'timeout': 10}})"
        )


DEVICE: Final = "/dev/ttyUSB0"


class TestSerialConnection:
    """Contains tests for SerialConnection class."""

    async def test_connect(
        self, serial_connection: SerialConnection, serial_asyncio_open_serial_connection
    ) -> None:
        """Test serial connect."""
        await serial_connection.connect()
        serial_asyncio_open_serial_connection.assert_called_once_with(
            url=DEVICE,
            baudrate=115200,
            bytesize=EIGHTBITS,
            parity=PARITY_NONE,
            stopbits=STOPBITS_ONE,
            timeout=10,
        )

    async def test_repr(self, serial_connection: SerialConnection) -> None:
        """Test serializable representation."""
        assert repr(serial_connection) == (
            "SerialConnection("
            f"device={DEVICE}, baudrate=115200, kwargs={{'timeout': 10}})"
        )



"""Contains frame classes."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cache, reduce
import struct
from typing import TYPE_CHECKING, Any, ClassVar, Final, TypeVar

from pyplumio.const import DeviceType, FrameType
from pyplumio.exceptions import UnknownFrameError
from pyplumio.helpers.factory import create_instance
from pyplumio.utils import ensure_dict, to_camelcase

FRAME_START: Final = 0x68
FRAME_END: Final = 0x16


HEADER_INDEX: Final = 0
FRAME_TYPE_SIZE: Final = 1
BCC_SIZE: Final = 1
BCC_INDEX: Final = -2
DELIMITER_SIZE: Final = 1

ECONET_TYPE: Final = 48
ECONET_VERSION: Final = 5

# Frame header structure.
struct_header = struct.Struct("<BH4B")
HEADER_SIZE = struct_header.size

if TYPE_CHECKING:
    from pyplumio.devices import PhysicalDevice


def bcc(buffer: bytes) -> int:
    """Return a block check character."""
    return reduce(lambda x, y: x ^ y, buffer)


@cache
def is_known_frame_type(frame_type: int) -> bool:
    """Check if frame type is known."""
    try:
        FrameType(frame_type)
        return True
    except ValueError:
        return False


@cache
def get_frame_handler(frame_type: int) -> str:
    """Return handler class path for the frame type."""
    if not is_known_frame_type(frame_type):
        raise UnknownFrameError(f"Unknown frame type ({frame_type})")

    module, type_name = FrameType(frame_type).name.split("_", 1)
    type_name = to_camelcase(type_name, overrides={"uid": "UID"})
    return f"frames.{module.lower()}s.{type_name}{module.capitalize()}"


@dataclass
class DataFrameDescription:
    """Describes what data is provided by the frame."""

    __slots__ = ("frame_type", "provides")

    frame_type: FrameType
    provides: str


FrameT = TypeVar("FrameT", bound="Frame")


class Frame(ABC):
    """Represents a frame."""

    __slots__ = (
        "recipient",
        "sender",
        "econet_type",
        "econet_version",
        "_handler",
        "_message",
        "_data",
    )

    recipient: DeviceType
    sender: DeviceType
    econet_type: int
    econet_version: int
    frame_type: ClassVar[FrameType]
    _handler: PhysicalDevice | None
    _message: bytearray | None
    _data: dict[str, Any] | None

    def __init__(
        self,
        recipient: DeviceType = DeviceType.ALL,
        sender: DeviceType = DeviceType.ECONET,
        econet_type: int = ECONET_TYPE,
        econet_version: int = ECONET_VERSION,
        message: bytearray | None = None,
        data: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Process a frame data and message."""
        self.recipient = recipient
        self.sender = sender
        self.econet_type = econet_type
        self.econet_version = econet_version
        self._handler = None
        self._data = data if not kwargs else ensure_dict(data, kwargs)
        self._message = message

    def __hash__(self) -> int:
        """Return a hash of the frame based on its values."""
        return hash(
            (
                self.recipient,
                self.sender,
                self.econet_type,
                self.econet_version,
                self._message,
                frozenset(self._data.items()) if self._data else None,
            )
        )

    def __eq__(self, other: object) -> bool:
        """Compare if this frame is equal to other."""
        if isinstance(other, Frame):
            return (
                self.recipient,
                self.sender,
                self.econet_type,
                self.econet_version,
                self._message,
                self._data,
            ) == (
                other.recipient,
                other.sender,
                other.econet_type,
                other.econet_version,
                other._message,
                other._data,
            )

        return NotImplemented

    def __repr__(self) -> str:
        """Return a serializable string representation."""
        return (
            f"{self.__class__.__name__}("
            f"recipient={repr(self.recipient)}, "
            f"sender={repr(self.sender)}, "
            f"econet_type={self.econet_type}, "
            f"econet_version={self.econet_version}, "
            f"message={self.message}, "
            f"data={self.data})"
        )

    def __len__(self) -> int:
        """Return a frame length."""
        return self.length

    def hex(self, *args: Any, **kwargs: Any) -> str:
        """Return a frame message represented as hex string."""
        return self.bytes.hex(*args, **kwargs)

    def assign_to(self, device: PhysicalDevice) -> None:
        """Assign device to the frame."""
        self._handler = device

    @property
    def handler(self) -> PhysicalDevice | None:
        """Return the device associated to the frame."""
        return self._handler

    @property
    def data(self) -> dict[str, Any]:
        """Return the frame data."""
        if self._data is None:
            self._data = (
                self.decode_message(self._message) if self._message is not None else {}
            )

        return self._data

    @data.setter
    def data(self, data: dict[str, Any]) -> None:
        """Set the frame data."""
        self._data = data
        self._message = None

    @property
    def message(self) -> bytearray:
        """Return the frame message."""
        if self._message is None:
            self._message = self.create_message(
                self._data if self._data is not None else {}
            )

        return self._message

    @message.setter
    def message(self, message: bytearray) -> None:
        """Set the frame message."""
        self._message = message
        self._data = None

    @property
    def length(self) -> int:
        """Return the frame length in bytes."""
        return (
            struct_header.size
            + FRAME_TYPE_SIZE
            + len(self.message)
            + BCC_SIZE
            + DELIMITER_SIZE
        )

    @property
    def header(self) -> bytearray:
        """Return the frame header."""
        buffer = bytearray(struct_header.size)
        struct_header.pack_into(
            buffer,
            HEADER_INDEX,
            FRAME_START,
            self.length,
            int(self.recipient),
            int(self.sender),
            self.econet_type,
            self.econet_version,
        )

        return buffer

    @property
    def bytes(self) -> bytes:
        """Return the frame bytes."""
        data = self.header
        data.append(self.frame_type)
        data += self.message
        data.append(bcc(data))
        data.append(FRAME_END)
        return bytes(data)

    @classmethod
    async def create(cls: type[FrameT], frame_type: int, **kwargs: Any) -> FrameT:
        """Create a frame handler object from frame type."""
        return await create_instance(get_frame_handler(frame_type), cls=cls, **kwargs)

    @abstractmethod
    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create frame message."""

    @abstractmethod
    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode frame message."""


class Request(Frame):
    """Represents a request."""

    __slots__ = ()

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        return bytearray()

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        return {}

    def response(self, **kwargs: Any) -> Response | None:
        """Return a response frame."""
        return None


class Response(Frame):
    """Represents a response."""

    __slots__ = ()

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        return bytearray()

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        return {}


class Message(Response):
    """Represents a message."""

    __slots__ = ()


__all__ = [
    "Frame",
    "Request",
    "Response",
    "Message",
    "DataFrameDescription",
    "bcc",
    "is_known_frame_type",
    "get_frame_handler",
]



"""Contains tests for the ecoMAX parameter descriptors."""

from pyplumio.const import ATTR_INDEX, ATTR_OFFSET, ATTR_SIZE, ATTR_VALUE, ProductType
from pyplumio.devices.ecomax import EcoMAX
from pyplumio.frames.requests import (
    EcomaxControlRequest,
    SetEcomaxParameterRequest,
    SetThermostatParameterRequest,
)
from pyplumio.parameters import ParameterValues
from pyplumio.parameters.ecomax import (
    PARAMETER_TYPES as ECOMAX_PARAMETER_TYPES,
    EcomaxNumber,
    EcomaxNumberDescription,
    EcomaxSwitch,
    EcomaxSwitchDescription,
    get_ecomax_parameter_types,
)
from pyplumio.structures.ecomax_parameters import ATTR_ECOMAX_CONTROL
from pyplumio.structures.product_info import ProductInfo
from pyplumio.structures.thermostat_parameters import ATTR_THERMOSTAT_PROFILE


async def test_ecomax_parameter_create_request(ecomax: EcoMAX) -> None:
    """Test create_request method for ecoMAX parameter."""
    # Test with ecoMAX control switch.
    ecomax_control = EcomaxSwitch(
        device=ecomax,
        description=EcomaxSwitchDescription(name=ATTR_ECOMAX_CONTROL),
        values=ParameterValues(value=0, min_value=0, max_value=1),
    )
    ecomax_control_request = await ecomax_control.create_request()
    assert isinstance(ecomax_control_request, EcomaxControlRequest)
    assert ecomax_control_request.data == {ATTR_VALUE: 0}

    # Test with thermostat profile number.
    thermostat_profile = EcomaxNumber(
        device=ecomax,
        description=EcomaxNumberDescription(name=ATTR_THERMOSTAT_PROFILE),
        values=ParameterValues(value=1, min_value=0, max_value=5),
    )
    thermostat_profile_request = await thermostat_profile.create_request()
    assert isinstance(thermostat_profile_request, SetThermostatParameterRequest)
    assert thermostat_profile_request.data == {
        ATTR_INDEX: 0,
        ATTR_VALUE: 1,
        ATTR_OFFSET: 0,
        ATTR_SIZE: 1,
    }

    # Test with generic number.
    ecomax_number = EcomaxNumber(
        device=ecomax,
        description=EcomaxNumberDescription(name="test_number"),
        values=ParameterValues(value=10, min_value=0, max_value=20),
    )
    ecomax_number_request = await ecomax_number.create_request()
    assert isinstance(ecomax_number_request, SetEcomaxParameterRequest)
    assert ecomax_number_request.data == {ATTR_INDEX: 0, ATTR_VALUE: 10}


async def test_get_ecomax_parameter_types(ecomax: EcoMAX) -> None:
    """Test ecoMAX parameter types getter."""
    parameter_types_all = ECOMAX_PARAMETER_TYPES
    assert len(parameter_types_all) == 2
    parameter_types = await get_ecomax_parameter_types(ecomax.product)
    assert len(parameter_types_all[ProductType.ECOMAX_P]) == 139
    assert len(parameter_types_all[ProductType.ECOMAX_I]) == 43
    assert parameter_types_all[ProductType.ECOMAX_P] == parameter_types

    # Test with patch.
    product_info = ProductInfo(
        type=ProductType.ECOMAX_P,
        id=48,
        uid="**REDACTED**",
        logo=48,
        image=2,
        model="ecoMAX 860D3-HB",
    )
    assert parameter_types[119].name == "water_heater_target_temp"
    assert parameter_types[120].name == "min_water_heater_target_temp"
    assert parameter_types[121].name == "max_water_heater_target_temp"
    assert parameter_types[127].name == "summer_mode_disable_temp"
    parameter_types_patched = await get_ecomax_parameter_types(product_info)
    assert parameter_types_patched[119].name == "summer_mode"
    assert parameter_types_patched[120].name == "summer_mode_enable_temp"
    assert parameter_types_patched[121].name == "summer_mode_disable_temp"
    assert parameter_types_patched[127].name == "__unknown_parameter_1"



"""Contains connection shortcuts and version information."""

from __future__ import annotations

from typing import Any

from pyplumio._version import __version__, __version_tuple__
from pyplumio.connection import SerialConnection, TcpConnection
from pyplumio.exceptions import (
    ChecksumError,
    ConnectionFailedError,
    FrameDataError,
    ProtocolError,
    PyPlumIOError,
    ReadError,
    RequestError,
    UnknownDeviceError,
    UnknownFrameError,
)
from pyplumio.frames import Frame
from pyplumio.protocol import AsyncProtocol, DummyProtocol, Protocol
from pyplumio.structures.network_info import EthernetParameters, WirelessParameters


def open_serial_connection(
    device: str,
    baudrate: int = 115200,
    *,
    protocol: Protocol | None = None,
    reconnect_on_failure: bool = True,
    **kwargs: Any,
) -> SerialConnection:
    r"""Create a serial connection.

    :param device: Serial port device name. e. g. /dev/ttyUSB0
    :type device: str
    :param baudrate: Serial port baud rate, defaults to 115200
    :type baudrate: int, optional
    :param protocol: Protocol that will be used for communication with
        the ecoMAX controller, default to AsyncProtocol
    :type protocol: Protocol, optional
    :param reconnect_on_failure: `True` if PyPlumIO should try
        reconnecting on failure, otherwise `False`, default to `True`
    :type reconnect_on_failure: bool, optional
    :param \**kwargs: Additional keyword arguments to be passed to
        serial_asyncio.open_serial_connection()
    :return: An instance of serial connection
    :rtype: SerialConnection
    """
    return SerialConnection(
        device,
        baudrate,
        protocol=protocol,
        reconnect_on_failure=reconnect_on_failure,
        **kwargs,
    )


def open_tcp_connection(
    host: str,
    port: int,
    *,
    protocol: Protocol | None = None,
    reconnect_on_failure: bool = True,
    **kwargs: Any,
) -> TcpConnection:
    r"""Create a TCP connection.

    :param host: IP address or host name of the remote RS-485 server
    :type host: str
    :param port: Port that remote RS-485 server is listening to
    :type port: int
    :param protocol: Protocol that will be used for communication with
        the ecoMAX controller, default to AsyncProtocol
    :type protocol: Protocol, optional
    :param reconnect_on_failure: `True` if PyPlumIO should try
        reconnecting on failure, otherwise `False`, default to `True`
    :type reconnect_on_failure: bool, optional
    :param \**kwargs: Additional keyword arguments to be passed to
        asyncio.open_connection()
    :return: An instance of TCP connection
    :rtype: TcpConnection
    """
    return TcpConnection(
        host,
        port,
        protocol=protocol,
        reconnect_on_failure=reconnect_on_failure,
        **kwargs,
    )


__all__ = [
    "__version__",
    "__version_tuple__",
    "AsyncProtocol",
    "ChecksumError",
    "ConnectionFailedError",
    "DummyProtocol",
    "EthernetParameters",
    "Frame",
    "FrameDataError",
    "Protocol",
    "ProtocolError",
    "PyPlumIOError",
    "ReadError",
    "RequestError",
    "SerialConnection",
    "TcpConnection",
    "UnknownDeviceError",
    "UnknownFrameError",
    "WirelessParameters",
    "open_serial_connection",
    "open_tcp_connection",
]



"""Contains a frame reader and writer classes."""

from __future__ import annotations

import asyncio
from asyncio import IncompleteReadError, StreamReader, StreamWriter
import logging
from typing import Final, NamedTuple, SupportsIndex

from pyplumio.const import DeviceType
from pyplumio.devices import is_known_device_type
from pyplumio.exceptions import ChecksumError, ReadError, UnknownDeviceError
from pyplumio.frames import (
    BCC_INDEX,
    DELIMITER_SIZE,
    FRAME_START,
    FRAME_TYPE_SIZE,
    HEADER_SIZE,
    Frame,
    bcc,
    struct_header,
)
from pyplumio.utils import timeout

READER_TIMEOUT: Final = 10
WRITER_TIMEOUT: Final = 10

MIN_FRAME_LENGTH: Final = 10
MAX_FRAME_LENGTH: Final = 1000

DEFAULT_BUFFER_SIZE: Final = 5000

_LOGGER = logging.getLogger(__name__)


class FrameWriter:
    """Represents a frame writer."""

    __slots__ = ("_writer",)

    _writer: StreamWriter

    def __init__(self, writer: StreamWriter) -> None:
        """Initialize a new frame writer."""
        self._writer = writer

    @timeout(WRITER_TIMEOUT)
    async def write(self, frame: Frame) -> None:
        """Send the frame and wait until send buffer is empty."""
        self._writer.write(frame.bytes)
        await self._writer.drain()
        _LOGGER.debug("Sent frame: %s, bytes: %s", frame, frame.bytes)

    async def close(self) -> None:
        """Close the frame writer."""
        try:
            self._writer.close()
            await self.wait_closed()
        except (OSError, asyncio.TimeoutError):
            _LOGGER.exception(
                "Failed to close the frame writer due to an unexpected error"
            )

    @timeout(WRITER_TIMEOUT)
    async def wait_closed(self) -> None:
        """Wait until the frame writer is closed."""
        await self._writer.wait_closed()


class BufferedReader:
    """Represents a buffered reader for reading frames."""

    __slots__ = ("_buffer", "_reader")

    _buffer: bytearray
    _reader: StreamReader

    def __init__(self, reader: StreamReader) -> None:
        """Initialize a new buffered reader."""
        self._buffer = bytearray()
        self._reader = reader

    async def ensure_buffer(self, size: int) -> None:
        """Ensure the internal buffer size."""
        bytes_to_read = size - len(self._buffer)
        if bytes_to_read <= 0:
            return None

        try:
            data = await self._reader.readexactly(bytes_to_read)
            self._buffer.extend(data)
            self.trim_to(DEFAULT_BUFFER_SIZE)
        except IncompleteReadError as e:
            raise ReadError(
                f"Incomplete read. Tried to read {bytes_to_read} additional bytes "
                f"to reach a total of {size}, but only {len(e.partial)} bytes were "
                "available from stream."
            ) from e
        except asyncio.CancelledError:
            _LOGGER.debug("Read operation cancelled while ensuring buffer")
            raise
        except Exception as e:
            raise OSError(
                f"Serial connection broken while trying to ensure {size} bytes: {e}"
            ) from e

    async def peek(self, size: int) -> bytes:
        """Read the specified number of bytes without consuming them."""
        await self.ensure_buffer(size)
        return memoryview(self._buffer)[:size].tobytes()

    async def consume(self, size: int) -> None:
        """Consume the specified number of bytes from the buffer."""
        await self.ensure_buffer(size)
        self._buffer = self._buffer[size:]

    async def read_into_buffer(self, size: int) -> None:
        """Read the specified number of bytes from the stream."""
        try:
            chunk = await self._reader.read(size)
        except asyncio.CancelledError:
            _LOGGER.debug("Read operation cancelled while filling internal buffer.")
            raise
        except Exception as e:
            raise OSError(
                f"Serial connection broken while filling internal buffer: {e}"
            ) from e

        if not chunk:
            _LOGGER.debug("Stream ended while filling internal buffer.")
            raise OSError(
                "Serial connection broken: stream ended while filling internal buffer"
            )

        self._buffer.extend(chunk)
        self.trim_to(DEFAULT_BUFFER_SIZE)

    def seek_to(self, delimiter: SupportsIndex) -> bool:
        """Trim the buffer to the first occurrence of the delimiter.

        Returns True if the delimiter was found and trimmed, False otherwise.
        """
        if not self._buffer or (index := self._buffer.find(delimiter)) == -1:
            return False

        self._buffer = self._buffer[index:]
        return True

    def trim_to(self, size: int) -> None:
        """Trim buffer to size."""
        if len(self._buffer) > size:
            self._buffer = self._buffer[-size:]

    @property
    def buffer(self) -> bytearray:
        """Return the internal buffer."""
        return self._buffer


class Header(NamedTuple):
    """Represents a frame header."""

    frame_length: int
    recipient: int
    sender: int
    econet_type: int
    econet_version: int


class FrameReader:
    """Represents a frame reader."""

    __slots__ = ("_reader",)

    _reader: BufferedReader

    def __init__(self, reader: StreamReader) -> None:
        """Initialize a new frame reader."""
        self._reader = BufferedReader(reader)

    async def _read_header(self) -> Header:
        """Locate and read a frame header."""
        while True:
            if self._reader.seek_to(FRAME_START):
                header_bytes = await self._reader.peek(HEADER_SIZE)
                header_fields = struct_header.unpack_from(header_bytes)
                return Header(*header_fields[DELIMITER_SIZE:])

            await self._reader.read_into_buffer(MAX_FRAME_LENGTH)

    @timeout(READER_TIMEOUT)
    async def read(self) -> Frame | None:
        """Read the frame and return corresponding handler object."""
        header = await self._read_header()
        frame_length, recipient, sender, econet_type, econet_version = header

        if frame_length > MAX_FRAME_LENGTH or frame_length < MIN_FRAME_LENGTH:
            await self._reader.consume(HEADER_SIZE)
            raise ReadError(
                f"Unexpected frame length ({frame_length}), expected between "
                f"{MIN_FRAME_LENGTH} and {MAX_FRAME_LENGTH}"
            )

        frame_bytes = await self._reader.peek(frame_length)
        checksum = bcc(frame_bytes[:BCC_INDEX])
        if checksum != frame_bytes[BCC_INDEX]:
            await self._reader.consume(HEADER_SIZE)
            raise ChecksumError(
                f"Incorrect frame checksum: calculated {checksum}, "
                f"expected {frame_bytes[BCC_INDEX]}. Frame data: {frame_bytes.hex()}"
            )

        await self._reader.consume(frame_length)
        if recipient not in (DeviceType.ECONET, DeviceType.ALL):
            _LOGGER.debug(
                "Skipping frame intended for different recipient (%s)", recipient
            )
            return None

        if not is_known_device_type(sender):
            raise UnknownDeviceError(f"Unknown sender type ({sender})")

        payload_bytes = frame_bytes[HEADER_SIZE:BCC_INDEX]
        frame = await Frame.create(
            frame_type=payload_bytes[0],
            recipient=DeviceType(recipient),
            sender=DeviceType(sender),
            econet_type=econet_type,
            econet_version=econet_version,
            message=payload_bytes[FRAME_TYPE_SIZE:],
        )
        _LOGGER.debug("Received frame: %s, bytes: %s", frame, frame_bytes.hex())

        return frame


__all__ = ["FrameReader", "FrameWriter"]



"""Contains tests for alerts structure decoder."""

import datetime as dt

import pytest

from pyplumio.const import AlertType
from pyplumio.frames.responses import AlertsResponse
from pyplumio.structures.alerts import (
    ATTR_ALERTS,
    ATTR_TOTAL_ALERTS,
    Alert,
    AlertsStructure,
    seconds_to_datetime,
)


@pytest.mark.parametrize(
    ("timestamp", "expected_datetime"),
    [
        (946684800, dt.datetime(2029, 6, 15, 0, 0, 0)),
        (1609459200, dt.datetime(2050, 1, 29, 0, 0)),
        (1704067200, dt.datetime(2053, 1, 8, 0, 0)),
        (1735689600, dt.datetime(2054, 1, 2, 0, 0)),
        (2147483647, dt.datetime(2066, 10, 25, 3, 14, 7)),
    ],
)
def test_seconds_to_datetime(timestamp: int, expected_datetime: dt.datetime) -> None:
    """Test conversion of seconds to datetime."""
    dt = seconds_to_datetime(timestamp)
    assert dt == expected_datetime


@pytest.fixture(name="alerts_structure")
def fixture_alerts_structure() -> AlertsStructure:
    """Fixture for AlertsStructure."""
    return AlertsStructure(frame=AlertsResponse())


class TestAlertStructure:
    """Test the Alert structure decoder."""

    def test_decode(self, alerts_structure: AlertsStructure) -> None:
        """Test decoding of alerts structure."""
        message = bytearray(
            [0x01, 0x00, 0x01, 0x00, 0xFF, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01]
        )
        data, offset = alerts_structure.decode(message)
        assert data == {
            ATTR_ALERTS: [
                Alert(
                    code=AlertType.POWER_LOSS,
                    from_dt=dt.datetime(2000, 7, 9, 22, 41, 3),
                    to_dt=dt.datetime(2000, 7, 9, 22, 36, 49),
                )
            ],
            ATTR_TOTAL_ALERTS: 1,
        }
        assert offset == 12

    def test_decode_without_alerts(self, alerts_structure: AlertsStructure) -> None:
        """Test decoding of alerts structure without any alerts."""
        message = bytearray([0x00, 0x00, 0x00])
        data, offset = alerts_structure.decode(message)
        assert data == {ATTR_TOTAL_ALERTS: 0}
        assert offset == 3



"""Contains tests for fuel level structure decoder."""


import pytest

from pyplumio.const import BYTE_UNDEFINED
from pyplumio.frames.messages import SensorDataMessage
from pyplumio.structures.fuel_level import ATTR_FUEL_LEVEL, FuelLevelStructure


@pytest.fixture(name="fuel_level_structure")
def fixture_fuel_level_structure() -> FuelLevelStructure:
    """Fixture for FuelLevelStructure."""
    return FuelLevelStructure(frame=SensorDataMessage())


class TestFuelLevelStructure:
    """Test the FuelLevel structure decoder."""

    def test_decode(self, fuel_level_structure: FuelLevelStructure) -> None:
        """Test decoding of fuel level structure."""
        message = bytearray([120])
        data, offset = fuel_level_structure.decode(message)
        assert data == {ATTR_FUEL_LEVEL: 19}
        assert offset == 1

    def test_decode_without_fuel_level(
        self, fuel_level_structure: FuelLevelStructure
    ) -> None:
        """Test decoding of fuel level structure without fuel level."""
        message = bytearray([BYTE_UNDEFINED])
        data, offset = fuel_level_structure.decode(message)
        assert data == {}
        assert offset == 1



"""Contains a mixer class."""

from __future__ import annotations

import asyncio
from collections.abc import Coroutine, Generator
import logging
from typing import Any

from pyplumio.devices import VirtualDevice
from pyplumio.helpers.event_manager import event_listener
from pyplumio.parameters import ParameterValues
from pyplumio.parameters.mixer import (
    MixerNumber,
    MixerSwitch,
    MixerSwitchDescription,
    get_mixer_parameter_types,
)
from pyplumio.structures.product_info import ATTR_PRODUCT, ProductInfo

_LOGGER = logging.getLogger(__name__)


class Mixer(VirtualDevice):
    """Represents a mixer."""

    __slots__ = ()

    @event_listener
    async def on_event_mixer_sensors(self, sensors: dict[str, Any]) -> bool:
        """Update mixer sensors and dispatch the events."""
        _LOGGER.debug("Received mixer %i sensors", self.index)
        await asyncio.gather(
            *(self.dispatch(name, value) for name, value in sensors.items())
        )
        return True

    @event_listener
    async def on_event_mixer_parameters(
        self, parameters: list[tuple[int, ParameterValues]]
    ) -> bool:
        """Update mixer parameters and dispatch the events."""
        _LOGGER.debug("Received mixer %i parameters", self.index)
        product_info: ProductInfo = await self.parent.get(ATTR_PRODUCT)
        parameter_types = get_mixer_parameter_types(product_info)

        def _mixer_parameter_events() -> Generator[Coroutine]:
            """Get dispatch calls for mixer parameter events."""
            for index, values in parameters:
                try:
                    description = parameter_types[index]
                except IndexError:
                    _LOGGER.warning(
                        "Encountered unknown mixer parameter (%i): %s. "
                        "Your device isn't fully compatible with this software "
                        "and may not work properly. Please visit the issue tracker "
                        "and open a feature request to support %s",
                        index,
                        values,
                        product_info.model,
                    )
                    return

                handler = (
                    MixerSwitch
                    if isinstance(description, MixerSwitchDescription)
                    else MixerNumber
                )
                yield self.dispatch(
                    description.name,
                    handler.create_or_update(
                        device=self, description=description, values=values, index=index
                    ),
                )

        await asyncio.gather(*_mixer_parameter_events())
        return True


__all__ = ["Mixer"]



"""Contains tests for the device handler classes."""

from __future__ import annotations

import asyncio
from typing import Literal
from unittest.mock import Mock, patch

import pytest

from pyplumio.const import ATTR_FRAME_ERRORS, DeviceType, FrameType
from pyplumio.devices import Device, PhysicalDevice, get_device_handler
from pyplumio.exceptions import RequestError, UnknownDeviceError
from pyplumio.filters import on_change
from pyplumio.frames import Response
from pyplumio.parameters import Parameter
from pyplumio.structures.frame_versions import ATTR_FRAME_VERSIONS
from pyplumio.structures.network_info import NetworkInfo
from tests.conftest import RAISES


@pytest.mark.parametrize(
    ("device_type", "handler"),
    [
        (DeviceType.ECOMAX, "devices.ecomax.EcoMAX"),
        (DeviceType.ECONET, "devices.econet.EcoNET"),
        (DeviceType.ECOSTER, "devices.ecoster.EcoSTER"),
        (99, RAISES),
    ],
)
def test_device_handler(
    device_type: DeviceType | int, handler: str | Literal["raises"]
) -> None:
    """Test getting device handler class by device address."""
    if handler == RAISES:
        with pytest.raises(UnknownDeviceError, match="Unknown device type"):
            get_device_handler(device_type)
    else:
        assert get_device_handler(device_type) == handler


class DummyDevice(Device):
    """Represents a dummy device."""


@pytest.fixture(name="device")
def fixture_device() -> Device:
    """Return a Device object."""
    return DummyDevice(asyncio.Queue())


class TestDevice:
    """Contains tests for the Device class."""

    @patch(
        "pyplumio.devices.Device.get", return_value=Mock(spec=Parameter, autospec=True)
    )
    async def test_set(self, mock_get, device: Device) -> None:
        """Test changing a device parameter."""
        await device.set("foo", "on")
        mock_get.assert_awaited_once_with("foo", None)
        parameter = mock_get.return_value
        parameter.set.assert_awaited_once_with("on", retries=0)

    @patch("pyplumio.devices.Device.get", return_value=Mock)
    async def test_set_incorrect_parameter(self, mock_get, device: Device) -> None:
        """Test changing an incorrect device parameter."""
        with pytest.raises(TypeError, match="not valid or does not exist"):
            await device.set("foo", "bar")  # type: ignore[arg-type]

    @patch("pyplumio.devices.Device.create_task")
    @patch("pyplumio.devices.Device.set", new_callable=Mock)
    def test_set_nowait(self, mock_set, mock_create_task, device: Device) -> None:
        """Test changing a device parameter without waiting."""
        device.set_nowait("foo", "off")
        mock_set.assert_called_once_with("foo", "off", 0, None)
        mock_create_task.assert_called_once_with(mock_set.return_value)

    @patch("pyplumio.devices.Device.cancel_tasks")
    @patch("pyplumio.devices.Device.wait_until_done")
    async def test_shutdown(
        self, wait_until_done, cancel_tasks, device: Device
    ) -> None:
        """Test shutting down the device tasks."""
        await device.shutdown()
        cancel_tasks.assert_called_once()
        wait_until_done.assert_awaited_once()


class DummyPhysicalDevice(PhysicalDevice):
    """Represents a dummy physical device."""

    address = 0


@pytest.fixture(name="physical_device")
def fixture_physical_device() -> PhysicalDevice:
    """Return a PhysicalDevice object."""
    return DummyPhysicalDevice(asyncio.Queue(), NetworkInfo())


class TestPhysicalDevice:
    """Contains tests for PhysicalDevice class."""

    @pytest.mark.parametrize(
        ("frame_type", "requested_frame_type"),
        [
            (FrameType.REQUEST_ALERTS, FrameType.REQUEST_ALERTS),
            (
                FrameType.REQUEST_ECOMAX_PARAMETER_CHANGES,
                FrameType.REQUEST_ECOMAX_PARAMETERS,
            ),
        ],
    )
    @patch("pyplumio.frames.Request.create", autospec=True)
    @patch("asyncio.Queue.put_nowait")
    async def test_frame_versions_event_listener(
        self,
        mock_put_nowait,
        mock_request_create,
        physical_device: PhysicalDevice,
        frame_type: FrameType,
        requested_frame_type: FrameType,
    ) -> None:
        """Test event listener for frame versions."""
        assert physical_device.has_frame_version(frame_type, 1) is False
        await physical_device.on_event_frame_versions({frame_type: 1})
        mock_request_create.assert_awaited_once_with(
            requested_frame_type, recipient=DummyPhysicalDevice.address
        )
        mock_put_nowait.assert_called_once_with(mock_request_create.return_value)
        assert physical_device.has_frame_version(frame_type, 1) is True

    def test_frame_versions_event_listener_decorator(self) -> None:
        """Test decorator for the frame version event listener."""
        event = getattr(PhysicalDevice.on_event_frame_versions, "_on_event", None)
        assert event == ATTR_FRAME_VERSIONS
        filter_func = getattr(
            PhysicalDevice.on_event_frame_versions, "_on_event_filter", None
        )
        assert filter_func is on_change

    async def test_supports_frame_type(self, physical_device: PhysicalDevice) -> None:
        """Test frame support checker."""
        assert physical_device.supports_frame_type(FrameType.REQUEST_ALERTS) is True
        physical_device.data = {ATTR_FRAME_ERRORS: [FrameType.REQUEST_ALERTS]}
        assert physical_device.supports_frame_type(FrameType.REQUEST_ALERTS) is False

    @patch("pyplumio.frames.Frame.assign_to")
    @patch("pyplumio.devices.PhysicalDevice.dispatch_nowait")
    async def test_handle_frame(
        self, mock_dispatch_nowait, mock_assign_to, physical_device: PhysicalDevice
    ) -> None:
        """Test frame handling."""
        frame = Response(data={"test": True})
        physical_device.handle_frame(frame)
        mock_assign_to.assert_called_once_with(physical_device)
        mock_dispatch_nowait.assert_called_once_with("test", True)

    @patch("pyplumio.devices.PhysicalDevice.get")
    @patch("pyplumio.frames.Request.create", autospec=True)
    @patch("asyncio.Queue.put_nowait")
    async def test_request(
        self,
        mock_put_nowait,
        mock_request_create,
        mock_get,
        physical_device: PhysicalDevice,
    ) -> None:
        """Test making a request."""
        await physical_device.request("alerts", frame_type=FrameType.REQUEST_ALERTS)
        mock_request_create.assert_called_once_with(
            FrameType.REQUEST_ALERTS, recipient=DummyPhysicalDevice.address
        )
        mock_put_nowait.assert_called_once_with(mock_request_create.return_value)
        mock_get.assert_awaited_once_with("alerts", timeout=3.0)

    @patch(
        "pyplumio.devices.PhysicalDevice.get",
        side_effect=(asyncio.TimeoutError, asyncio.TimeoutError),
    )
    @patch("pyplumio.frames.Request.create", autospec=True)
    @patch("asyncio.Queue.put_nowait")
    async def test_request_retry(
        self,
        mock_put_nowait,
        mock_request_create,
        mock_get,
        physical_device: PhysicalDevice,
    ) -> None:
        """Test retrying a request."""
        with pytest.raises(RequestError, match="Failed to request"):
            await physical_device.request(
                "alerts", frame_type=FrameType.REQUEST_ALERTS, retries=2
            )

        mock_request_create.assert_called_once_with(
            FrameType.REQUEST_ALERTS, recipient=DummyPhysicalDevice.address
        )
        mock_put_nowait.assert_called_with(mock_request_create.return_value)
        assert mock_put_nowait.call_count == 2
        mock_get.assert_awaited_with("alerts", timeout=3.0)
        assert mock_get.call_count == 2



"""Contains fixtures for the test suite."""

from __future__ import annotations

import asyncio
from collections.abc import Sequence
import functools
import importlib
import inspect
import json
from math import isclose
import os
import pathlib
from typing import Any, Final, TypeVar
from unittest.mock import AsyncMock, patch

from freezegun import freeze_time
import pytest

from pyplumio.const import ProductType, State
from pyplumio.devices.ecomax import EcoMAX
from pyplumio.parameters import NumericType
from pyplumio.structures.network_info import NetworkInfo
from pyplumio.structures.product_info import ATTR_PRODUCT, ProductInfo

TESTDATA_DIR: Final = "testdata"
UNDEFINED: Final = "undefined"
RAISES: Final = "raises"
DEFAULT_TOLERANCE: Final = 1e-6

T = TypeVar("T")


def _create_class_instance(module_name: str, class_name: str, **kwargs):
    """Create class instance and cache it."""
    return getattr(importlib.import_module(module_name), class_name)(**kwargs)


def _try_int(key: Any) -> Any:
    """Try to convert key to integer or return key unchanged on error."""
    try:
        return int(key)
    except ValueError:
        return key


def _decode_hinted_objects(d: Any) -> Any:
    """Decode a hinted JSON objects."""
    if "__module__" in d and "__class__" in d:
        module_name = d.pop("__module__")
        class_name = d.pop("__class__")
        return _create_class_instance(module_name, class_name, **d)

    if "__bytearray__" in d:
        return bytearray.fromhex("".join(d["items"]))

    if "__tuple__" in d:
        return tuple(d["items"])

    return {_try_int(k): v for k, v in d.items()}


def load_json_test_data(path: str) -> Any:
    """Load test data from JSON file."""
    abs_path = "/".join([os.path.dirname(__file__), TESTDATA_DIR, path])
    file = pathlib.Path(abs_path)
    with open(file, encoding="utf-8") as fp:
        return json.load(fp, object_hook=_decode_hinted_objects)


def load_json_parameters(path: str):
    """Prepare JSON test data for parametrization."""
    test_data = load_json_test_data(path)
    return [pytest.param(x["message"], x["data"], id=x["id"]) for x in test_data]


def _bypass_pytest_argument_inspection(
    wrapper: T, signature: inspect.Signature, argument: str
) -> T:
    """Remove argument from pytest's parameter inspection."""
    replacement = signature.replace(
        parameters=[p for p in signature.parameters.values() if p.name != argument]
    )
    setattr(wrapper, "__signature__", replacement)
    return wrapper


def json_test_data(
    json_path: str,
    selector: str | None = None,
    dataset: int = 0,
    name: str | None = None,
):
    """Pytest decorator to inject JSON test data as a test argument."""
    if not name:
        name = json_path.split("/")[-1].split("\\")[-1].rsplit(".", 1)[0]

    if selector:
        name += f"_{selector}"

    def decorator(test_func):
        @functools.wraps(test_func)
        async def wrapper(*args, **kwargs):
            json_test_data = load_json_test_data(json_path)[dataset]
            kwargs[name] = json_test_data[selector] if selector else json_test_data
            return await test_func(*args, **kwargs)

        return _bypass_pytest_argument_inspection(
            wrapper, inspect.signature(test_func), name
        )

    return decorator


def class_from_json(
    cls: type,
    json_path: str,
    /,
    dataset: int = 0,
    arguments: Sequence | None = None,
    name: str | None = None,
):
    """Pytest decorator to inject JSON test data as a test argument."""
    if not name:
        name = json_path.split("/")[-1].split("\\")[-1].rsplit(".", 1)[0]

    init_args = arguments if arguments else ()

    def decorator(test_func):
        @functools.wraps(test_func)
        async def wrapper(*args, **kwargs):
            json_test_data = load_json_test_data(json_path)[dataset]
            kwargs[name] = cls(**{k: json_test_data[k] for k in init_args})
            return await test_func(*args, **kwargs)

        return _bypass_pytest_argument_inspection(
            wrapper, inspect.signature(test_func), name
        )

    return decorator


def equal_parameter_value(
    a: NumericType | State | None, b: NumericType | State | None
) -> bool:
    """Compare the parameter values."""
    if isinstance(a, float) and isinstance(b, float):
        return isclose(a, b, rel_tol=DEFAULT_TOLERANCE)
    else:
        return True if a == b else False


@pytest.fixture(name="skip_asyncio_events")
def fixture_skip_asyncio_events():
    """Bypass asyncio events."""
    with (
        patch("asyncio.Event.wait", new_callable=AsyncMock),
        patch("asyncio.Event.is_set", return_value=True),
    ):
        yield


@pytest.fixture(autouse=True)
def skip_asyncio_sleep():
    """Skip an asyncio sleep calls."""
    with patch("asyncio.sleep"):
        yield


@pytest.fixture(name="ecomax")
def fixture_ecomax() -> EcoMAX:
    """Return an ecoMAX object."""
    ecomax = EcoMAX(asyncio.Queue(), network=NetworkInfo())
    ecomax.data[ATTR_PRODUCT] = ProductInfo(
        type=ProductType.ECOMAX_P,
        id=90,
        uid="TEST",
        logo=23040,
        image=2816,
        model="ecoMAX 350P2-ZF",
    )
    return ecomax


@pytest.fixture(name="frozen_time")
def fixture_frozen_time():
    """Get frozen time."""
    with freeze_time("2012-12-12 12:00:00") as frozen_time:
        yield frozen_time


__all__ = [
    "class_from_json",
    "equal_parameter_value",
    "json_test_data",
    "load_json_parameters",
    "load_json_test_data",
]



"""Contains a mixer parameter structure decoder."""

from __future__ import annotations

from collections.abc import Generator
from typing import Any, Final

from pyplumio.parameters import ParameterValues, unpack_parameter
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_MIXER_PARAMETERS: Final = "mixer_parameters"

MIXER_PARAMETER_SIZE: Final = 3


class MixerParametersStructure(StructureDecoder):
    """Represents a mixer parameters data structure."""

    __slots__ = ("_offset",)

    _offset: int

    def _mixer_parameter(
        self, message: bytearray, start: int, end: int
    ) -> Generator[tuple[int, ParameterValues], None, None]:
        """Get a single mixer parameter."""
        for index in range(start, start + end):
            if parameter := unpack_parameter(message, self._offset):
                yield (index, parameter)

            self._offset += MIXER_PARAMETER_SIZE

    def _mixer_parameters(
        self, message: bytearray, mixers: int, start: int, end: int
    ) -> Generator[tuple[int, list[tuple[int, ParameterValues]]], None, None]:
        """Get parameters for a mixer."""
        for index in range(mixers):
            if parameters := list(self._mixer_parameter(message, start, end)):
                yield (index, parameters)

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        start = message[offset + 1]
        end = message[offset + 2]
        mixers = message[offset + 3]
        self._offset = offset + 4
        return (
            ensure_dict(
                data,
                {
                    ATTR_MIXER_PARAMETERS: dict(
                        self._mixer_parameters(message, mixers, start, end)
                    )
                },
            ),
            self._offset,
        )


__all__ = [
    "ATTR_MIXER_PARAMETERS",
    "MixerParametersStructure",
]



"""Contains tests for the filter classes."""

from datetime import datetime, timedelta
from importlib import reload
import logging
import sys
from unittest.mock import AsyncMock, Mock, patch

import pytest

import pyplumio
from pyplumio import filters
import pyplumio.filters
from pyplumio.parameters import Parameter, ParameterValues
from pyplumio.structures.alerts import Alert


@pytest.fixture(name="use_numpy", params=(True, False))
def fixture_use_numpy(request, monkeypatch, caplog):
    """Try with and without numpy package."""
    if not request.param:
        monkeypatch.setitem(sys.modules, "numpy", None)

    with caplog.at_level(logging.INFO):
        reload(pyplumio.filters)

    message = "Using numpy for improved float precision"
    if request.param:
        assert message in caplog.text
    else:
        assert message not in caplog.text

    return request.param


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        (1, 10),
        (50, 15),
        (11, 11),
    ],
)
async def test_clamp(input_value, expected) -> None:
    """Test the clamp filter."""
    test_callback = AsyncMock()
    wrapped_callback = filters.clamp(test_callback, min_value=10, max_value=15)
    assert hash(wrapped_callback) == hash(test_callback)
    await wrapped_callback(input_value)
    test_callback.assert_awaited_once_with(expected)


async def test_deadband() -> None:
    """Test the deadband filter."""
    test_callback = AsyncMock()
    wrapped_callback = filters.deadband(test_callback, tolerance=0.1)
    assert hash(wrapped_callback) == hash(test_callback)
    await wrapped_callback(1)
    test_callback.assert_awaited_once_with(1)
    test_callback.reset_mock()

    # Check that callback is not awaited on insignificant change.
    await wrapped_callback(1.01)
    test_callback.assert_not_awaited()

    await wrapped_callback(1.1)
    test_callback.assert_awaited_once_with(1.1)

    # Test with non-numeric value.
    with pytest.raises(TypeError, match="filter can only be used with numeric values"):
        await wrapped_callback("banana")


async def test_on_change() -> None:
    """Test the value changed filter."""
    test_callback = AsyncMock()
    wrapped_callback = filters.on_change(test_callback)
    assert hash(wrapped_callback) == hash(test_callback)
    await wrapped_callback(1)
    test_callback.assert_awaited_once_with(1)
    test_callback.reset_mock()

    await wrapped_callback(1.1)
    test_callback.assert_awaited_once_with(1.1)

    # Test equality with callback function and different instance.
    assert test_callback == wrapped_callback
    assert wrapped_callback == filters.on_change(test_callback)
    assert wrapped_callback.__eq__("you shall not pass") is NotImplemented


async def test_on_change_parameter() -> None:
    """Test the value changed filter with parameters."""
    test_callback = AsyncMock()
    test_parameter = AsyncMock(spec=Parameter)
    test_parameter.values = ParameterValues(0, 0, 1)
    test_parameter.update_pending.is_set = Mock(return_value=False)
    wrapped_callback = filters.on_change(test_callback)
    assert hash(wrapped_callback) == hash(test_callback)
    await wrapped_callback(test_parameter)
    test_callback.assert_awaited_once_with(test_parameter)
    test_callback.reset_mock()

    # Check that callback is not awaited with no change.
    await wrapped_callback(test_parameter)
    test_callback.assert_not_awaited()

    # Check that callback is awaited on local value change.
    test_parameter.update_pending.is_set.return_value = True
    await wrapped_callback(test_parameter)
    test_callback.assert_awaited_once_with(test_parameter)
    test_callback.reset_mock()
    test_parameter.update_pending.is_set.return_value = False

    # Check that callback is awaited on remote value change.
    test_parameter.values = ParameterValues(1, 0, 1)
    await wrapped_callback(test_parameter)
    test_callback.assert_awaited_once_with(test_parameter)
    test_callback.reset_mock()

    # Check that callback is awaited on min value change.
    test_parameter.values = ParameterValues(1, 1, 1)
    await wrapped_callback(test_parameter)
    test_callback.assert_awaited_once_with(test_parameter)
    test_callback.reset_mock()

    # Check that callback is awaited on max value change.
    test_parameter.values = ParameterValues(1, 1, 2)
    await wrapped_callback(test_parameter)
    test_callback.assert_awaited_once_with(test_parameter)


async def test_debounce() -> None:
    """Test the debounce filter."""
    test_callback = AsyncMock()
    wrapped_callback = filters.debounce(test_callback, min_calls=3)
    assert hash(wrapped_callback) == hash(test_callback)
    await wrapped_callback(1)
    test_callback.assert_awaited_once_with(1)
    test_callback.reset_mock()

    # Ignore stray "1" and only await callback on a "2".
    await wrapped_callback(2)
    await wrapped_callback(1)
    await wrapped_callback(2)
    await wrapped_callback(2)
    await wrapped_callback(2)
    test_callback.assert_awaited_once_with(2)


async def test_throttle(frozen_time) -> None:
    """Test the throttle filter."""
    test_callback = AsyncMock()
    wrapped_callback = filters.throttle(test_callback, seconds=5)
    assert hash(wrapped_callback) == hash(test_callback)
    await wrapped_callback(1)
    test_callback.assert_awaited_once_with(1)
    test_callback.reset_mock()

    # One second passed.
    frozen_time.tick(timedelta(seconds=1))
    await wrapped_callback(2)
    test_callback.assert_not_awaited()

    # Five seconds passed.
    frozen_time.tick(timedelta(seconds=4))
    await wrapped_callback(3)
    test_callback.assert_awaited_once_with(3)
    test_callback.reset_mock()

    # Six seconds passed.
    frozen_time.tick(timedelta(seconds=1))
    await wrapped_callback(4)
    test_callback.assert_not_awaited()


async def test_delta() -> None:
    """Test the delta filter."""
    test_callback = AsyncMock()
    wrapped_callback = filters.delta(test_callback)
    assert hash(wrapped_callback) == hash(test_callback)

    await wrapped_callback(5)
    test_callback.assert_not_awaited()

    await wrapped_callback(3)
    test_callback.assert_awaited_once_with(-2)
    test_callback.reset_mock()

    # Test with list of alerts.
    alert1 = Alert(code=0, from_dt=datetime.now(), to_dt=None)
    alert2 = Alert(code=1, from_dt=datetime.now(), to_dt=None)
    alert3 = Alert(code=2, from_dt=datetime.now(), to_dt=None)
    wrapped_callback = filters.delta(test_callback)
    await wrapped_callback([alert1, alert2])
    await wrapped_callback([alert3, alert2])
    test_callback.assert_awaited_once_with([alert3])
    test_callback.reset_mock()

    # Test with unknown.
    wrapped_callback = filters.delta(test_callback)
    await wrapped_callback("foo")
    await wrapped_callback("bar")
    test_callback.assert_not_awaited()


async def test_aggregate(use_numpy, frozen_time) -> None:
    """Test the aggregate filter."""
    test_callback = AsyncMock()
    wrapped_callback = filters.aggregate(test_callback, seconds=5, sample_size=5)
    assert hash(wrapped_callback) == hash(test_callback)
    await wrapped_callback(1)
    test_callback.assert_not_awaited()

    # One second passed.
    frozen_time.tick(timedelta(seconds=1))
    await wrapped_callback(1)
    test_callback.assert_not_awaited()

    # Five seconds passed.
    frozen_time.tick(timedelta(seconds=4))
    if use_numpy:
        with patch("numpy.sum", return_value=5) as mock_sum:
            await wrapped_callback(3)

        mock_sum.assert_called_once_with([1, 1, 3])
    else:
        await wrapped_callback(3)

    test_callback.assert_awaited_once_with(5)
    test_callback.reset_mock()

    # Six second passed.
    frozen_time.tick(timedelta(seconds=1))
    await wrapped_callback(3)
    test_callback.assert_not_awaited()

    # Test with non-numeric value.
    with pytest.raises(TypeError, match="filter can only be used with numeric values"):
        await wrapped_callback("banana")


async def test_aggregate_sample_size(frozen_time) -> None:
    """Test the aggregate filter with sample size."""
    test_callback = AsyncMock()
    wrapped_callback = filters.aggregate(test_callback, seconds=5, sample_size=2)
    assert hash(wrapped_callback) == hash(test_callback)

    # Zero seconds passed, current sample size is 1.
    await wrapped_callback(1)
    test_callback.assert_not_awaited()

    # One second passed, current sample size is 2.
    frozen_time.tick(timedelta(seconds=1))
    await wrapped_callback(1)
    test_callback.assert_awaited_once_with(2)

    # Two seconds passed, current sample size is 3.
    frozen_time.tick(timedelta(seconds=1))
    await wrapped_callback(3)
    test_callback.test_callback.assert_not_awaited()


@pytest.mark.parametrize(
    ("filter_func", "input_value", "callback"),
    [
        (lambda x: len(x) == 4, [1, 2], False),
        (lambda x: len(x) == 4, [1, 2, 3, 4], True),
        (lambda x: len(x) == 4, [], False),
    ],
)
async def test_custom(filter_func, input_value, callback) -> None:
    """Test the custom filter."""
    test_callback = AsyncMock()
    wrapped_callback = filters.custom(test_callback, filter_func)
    assert hash(wrapped_callback) == hash(test_callback)
    await wrapped_callback(input_value)

    if callback:
        test_callback.assert_awaited_once_with(input_value)
    else:
        test_callback.assert_not_awaited()



"""Test boiler load structure decoder."""

import pytest

from pyplumio.const import BYTE_UNDEFINED
from pyplumio.frames.messages import SensorDataMessage
from pyplumio.structures.boiler_load import ATTR_BOILER_LOAD, BoilerLoadStructure


@pytest.fixture(name="boiler_load_structure")
def fixture_boiler_load_structure() -> BoilerLoadStructure:
    """Fixture for AlertsStructure."""
    return BoilerLoadStructure(frame=SensorDataMessage())


class TestBoilerLoadStructure:
    """Test the BoilerLoad structure decoder."""

    def test_decode(self, boiler_load_structure: BoilerLoadStructure) -> None:
        """Test decoding of boiler load structure."""
        message = bytearray([50])
        data, offset = boiler_load_structure.decode(message)
        assert data == {ATTR_BOILER_LOAD: 50}
        assert offset == 1

    def test_decode_without_boiler_load(
        self, boiler_load_structure: BoilerLoadStructure
    ) -> None:
        """Test decoding of boiler load structure without boiler load."""
        message = bytearray([BYTE_UNDEFINED])
        data, offset = boiler_load_structure.decode(message)
        assert data == {}
        assert offset == 1



"""Contains tests for the response frame classes."""

import pytest
from tests.conftest import load_json_parameters

from pyplumio.devices.ecomax import EcoMAX
from pyplumio.frames.responses import (
    AlertsResponse,
    DeviceAvailableResponse,
    EcomaxParametersResponse,
    MixerParametersResponse,
    PasswordResponse,
    ProgramVersionResponse,
    RegulatorDataSchemaResponse,
    SchedulesResponse,
    ThermostatParametersResponse,
    UIDResponse,
)
from pyplumio.structures.thermostat_sensors import ATTR_THERMOSTATS_AVAILABLE


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("responses/alerts.json"),
)
def test_alerts_response(message, data) -> None:
    """Test an alerts response."""
    assert AlertsResponse(message=message).data == data
    assert not AlertsResponse(data=data).message


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("responses/regulator_data_schema.json"),
)
def test_regulator_data_schema_response(message, data) -> None:
    """Test a regulator data schema response."""
    assert RegulatorDataSchemaResponse(message=message).data == data
    assert not RegulatorDataSchemaResponse(data=data).message


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("responses/device_available.json"),
)
def test_device_available_response(message, data) -> None:
    """Test a device available response."""
    assert DeviceAvailableResponse(data=data).message == message
    assert DeviceAvailableResponse(message=message).data == data


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("responses/ecomax_parameters.json"),
)
def test_ecomax_parameters_response(message, data) -> None:
    """Test a ecoMAX parameters response."""
    assert EcomaxParametersResponse(message=message).data == data
    assert not EcomaxParametersResponse(data=data).message


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("responses/mixer_parameters.json"),
)
def test_mixer_parameters_response(message, data) -> None:
    """Test a mixer parameters response."""
    assert MixerParametersResponse(message=message).data == data
    assert not MixerParametersResponse(data=data).message


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("responses/password.json"),
)
def test_password_response(message, data) -> None:
    """Test parsing password message."""
    assert PasswordResponse(message=message).data == data
    assert not PasswordResponse(data=data).message


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("responses/program_version.json"),
)
def test_program_version_response(message, data) -> None:
    """Test a program version response."""
    assert ProgramVersionResponse(data=data).message == message
    assert ProgramVersionResponse(message=message).data == data


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("responses/schedules.json"),
)
def test_schedules_response(message, data) -> None:
    """Test a schedules response."""
    assert SchedulesResponse(message=message).data == data
    assert not SchedulesResponse(data=data).message


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("responses/thermostat_parameters.json"),
)
async def test_thermostat_parameters_response(ecomax: EcoMAX, message, data) -> None:
    """Test a thermostat parameters response."""
    frame = ThermostatParametersResponse(message=message)
    frame.assign_to(ecomax)
    ecomax.load_nowait({ATTR_THERMOSTATS_AVAILABLE: 3})
    await ecomax.wait_until_done()
    assert frame.data == data


@pytest.mark.parametrize(
    ("message", "data"),
    load_json_parameters("responses/uid.json"),
)
def test_uid_response(message, data) -> None:
    """Test an UID response."""
    assert UIDResponse(message=message).data == data
    assert not UIDResponse(data=data).message



"""Contains tests for frame versions structure decoder."""

import pytest

from pyplumio.const import FrameType
from pyplumio.frames.messages import SensorDataMessage
from pyplumio.structures.frame_versions import (
    ATTR_FRAME_VERSIONS,
    FrameVersionsStructure,
)


@pytest.fixture(name="frame_versions_structure")
def fixture_frame_versions_structure() -> FrameVersionsStructure:
    """Fixture for BoilerPowerStructure."""
    return FrameVersionsStructure(frame=SensorDataMessage())


class TestFrameVersionsStructure:
    """Test the FrameVersions structure decoder."""

    def test_decode(self, frame_versions_structure: FrameVersionsStructure) -> None:
        """Test decoding of frame versions structure."""
        message = bytearray([0x01, 0x31, 0xFF, 0x00, 0x00, 0x00])
        data, offset = frame_versions_structure.decode(message)
        assert data == {ATTR_FRAME_VERSIONS: {FrameType.REQUEST_ECOMAX_PARAMETERS: 255}}
        assert offset == 4



"""Contains tests for the product info structure decoder."""

import pytest

from pyplumio.structures.product_info import format_model_name, unpack_uid


@pytest.mark.parametrize(
    ("message", "uid"),
    [
        (bytearray.fromhex("001600110D383338365539"), "D251PAKR3GCPZ1K8G05G0"),
        (bytearray.fromhex("002500300E191932135831"), "CE71HB09J468P1ZZ00980"),
    ],
)
def test_from_bytes(message: bytearray, uid: str) -> None:
    """Test unpacking an UID from bytes."""
    assert unpack_uid(message) == uid


@pytest.mark.parametrize(
    ("model_name", "formatted_name"),
    [
        ("EM360P2-ZF", "ecoMAX 360P2-ZF"),
        ("ecoMAXX800R3", "ecoMAXX 800R3"),
        ("ecoMAX850P2-C", "ecoMAX 850P2-C"),
        ("ecoMAX 850i", "ecoMAX 850i"),
        ("ecoMAX 860D3-HB", "ecoMAX 860D3-HB"),
        ("UNKNOWN", "UNKNOWN"),
    ],
)
def test_format_model_name(model_name: str, formatted_name: str) -> None:
    """Test formatting of model names."""
    assert format_model_name(model_name) == formatted_name



"""Contains an alerts structure decoder."""

from __future__ import annotations

from collections.abc import Generator
from contextlib import suppress
from dataclasses import dataclass
from datetime import datetime
from functools import lru_cache
from typing import Any, Final, Literal, NamedTuple

from pyplumio.const import AlertType
from pyplumio.data_types import UnsignedInt
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_ALERTS: Final = "alerts"
ATTR_TOTAL_ALERTS: Final = "total_alerts"

MAX_UINT32: Final = 0xFFFFFFFF


class DateTimeInterval(NamedTuple):
    """Represents an alert time interval."""

    name: Literal["year", "month", "day", "hour", "minute", "second"]
    seconds: int
    offset: int = 0


DATETIME_INTERVALS: tuple[DateTimeInterval, ...] = (
    DateTimeInterval("year", seconds=60 * 60 * 24 * 31 * 12, offset=2000),
    DateTimeInterval("month", seconds=60 * 60 * 24 * 31, offset=1),
    DateTimeInterval("day", seconds=60 * 60 * 24, offset=1),
    DateTimeInterval("hour", seconds=60 * 60),
    DateTimeInterval("minute", seconds=60),
    DateTimeInterval("second", seconds=1),
)


@lru_cache(maxsize=10)
def seconds_to_datetime(timestamp: int) -> datetime:
    """Convert timestamp to a datetime object.

    The ecoMAX controller stores alert time as a special timestamp value
    in seconds counted from Jan 1st, 2000.
    """

    def datetime_kwargs(timestamp: int) -> Generator[Any, None, None]:
        """Yield a tuple, that represents a single datetime kwarg."""
        for name, seconds, offset in DATETIME_INTERVALS:
            value = timestamp // seconds
            timestamp -= value * seconds
            yield name, (value + offset)

    return datetime(**dict(datetime_kwargs(timestamp)))


@dataclass
class Alert:
    """Represents a device alert."""

    __slots__ = ("code", "from_dt", "to_dt")

    code: int
    from_dt: datetime
    to_dt: datetime | None


class AlertsStructure(StructureDecoder):
    """Represents an alerts data structure."""

    __slots__ = ("_offset",)

    _offset: int

    def _unpack_alert(self, message: bytearray) -> Alert:
        """Unpack an alert."""
        offset = self._offset
        code = message[offset]
        offset += 1
        from_seconds = UnsignedInt.from_bytes(message, offset)
        offset += from_seconds.size
        to_seconds = UnsignedInt.from_bytes(message, offset)
        offset += to_seconds.size
        from_dt = seconds_to_datetime(from_seconds.value)
        to_dt = (
            None
            if to_seconds.value == MAX_UINT32
            else seconds_to_datetime(to_seconds.value)
        )
        with suppress(ValueError):
            code = AlertType(code)

        self._offset = offset
        return Alert(code, from_dt, to_dt)

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        total_alerts = message[offset]
        start = message[offset + 1]
        end = message[offset + 2]
        self._offset = offset + 3
        if end == 0:
            # No alerts found.
            return ensure_dict(data, {ATTR_TOTAL_ALERTS: total_alerts}), self._offset

        return (
            ensure_dict(
                data,
                {
                    ATTR_ALERTS: [
                        self._unpack_alert(message) for _ in range(start, start + end)
                    ],
                    ATTR_TOTAL_ALERTS: total_alerts,
                },
            ),
            self._offset,
        )


__all__ = ["AlertsStructure", "Alert", "ATTR_ALERTS", "ATTR_TOTAL_ALERTS"]



"""Contains a simple async cache for caching results of async functions."""

from collections.abc import Awaitable
from functools import wraps
from typing import Any, Callable, TypeVar, cast

from typing_extensions import ParamSpec, TypeAlias

T = TypeVar("T")
P = ParamSpec("P")
_CallableT: TypeAlias = Callable[..., Awaitable[Any]]


class AsyncCache:
    """A simple cache for asynchronous functions."""

    __slots__ = ("cache",)

    cache: dict[str, Any]

    def __init__(self) -> None:
        """Initialize the cache."""
        self.cache = {}

    async def get(self, key: str, coro: _CallableT) -> Any:
        """Get a value from the cache or compute and store it."""
        if key not in self.cache:
            self.cache[key] = await coro()

        return self.cache[key]


# Create a global cache instance
async_cache = AsyncCache()


def acache(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
    """Cache the result of an async function."""

    @wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        func_name = f"{func.__module__}.{func.__qualname__}"
        key = f"{func_name}:{args}:{kwargs}"
        return cast(T, await async_cache.get(key, lambda: func(*args, **kwargs)))

    return wrapper


__all__ = ["acache", "async_cache"]



"""Test lambda sensor structure decoder."""

import pytest

from pyplumio.const import BYTE_UNDEFINED, LambdaState
from pyplumio.frames.messages import SensorDataMessage
from pyplumio.structures.lambda_sensor import (
    ATTR_LAMBDA_LEVEL,
    ATTR_LAMBDA_STATE,
    ATTR_LAMBDA_TARGET,
    LambdaSensorStructure,
)


@pytest.fixture(name="lambda_sensor_structure")
def fixture_lambda_sensor_structure() -> LambdaSensorStructure:
    """Fixture for LambdaSensorStructure."""
    return LambdaSensorStructure(frame=SensorDataMessage())


class TestLambdaSensorStructure:
    """Test the LambdaSensor structure decoder."""

    def test_decode(self, lambda_sensor_structure: LambdaSensorStructure) -> None:
        """Test decoding of lambda sensor structure."""
        message = bytearray([0x00, 0x01, 0x4B, 0x00])
        data, offset = lambda_sensor_structure.decode(message)
        assert data == {
            ATTR_LAMBDA_STATE: LambdaState.STOP,
            ATTR_LAMBDA_TARGET: 1,
            ATTR_LAMBDA_LEVEL: 7.5,
        }
        assert offset == 4

    def test_decode_without_lambda_sensor(
        self, lambda_sensor_structure: LambdaSensorStructure
    ) -> None:
        """Test decoding of lambda sensor structure without lambda sensor."""
        message = bytearray([BYTE_UNDEFINED])
        data, offset = lambda_sensor_structure.decode(message)
        assert data == {}
        assert offset == 1



"""Contains tests for the schedules structure decoder."""

from __future__ import annotations

import asyncio
import datetime as dt
from typing import Literal, cast
from unittest.mock import Mock, patch

import pytest

from pyplumio.const import (
    ATTR_PARAMETER,
    ATTR_SCHEDULE,
    ATTR_SWITCH,
    ATTR_TYPE,
    STATE_OFF,
    STATE_ON,
    FrameType,
    State,
)
from pyplumio.devices import DeviceType, PhysicalDevice
from pyplumio.devices.ecomax import EcoMAX
from pyplumio.structures.schedules import (
    ATTR_SCHEDULE_PARAMETER,
    ATTR_SCHEDULE_SWITCH,
    ATTR_SCHEDULES,
    MIDNIGHT,
    MIDNIGHT_DT,
    STEP,
    TIME_FORMAT,
    Schedule,
    ScheduleDay,
    Time,
    collect_schedule_data,
    get_time,
    get_time_range,
)
from tests.conftest import RAISES


def test_collect_schedule_data(ecomax: EcoMAX) -> None:
    """Test collecting schedule data."""
    schedule_name = "heating"
    ecomax.data |= {
        f"{schedule_name}_{ATTR_SCHEDULE_SWITCH}": 1,
        f"{schedule_name}_{ATTR_SCHEDULE_PARAMETER}": 2,
        ATTR_SCHEDULES: {schedule_name: {"test": 1}},
    }
    data = collect_schedule_data(schedule_name, ecomax)
    assert data == {
        "parameter": 2,
        "schedule": {"test": 1},
        "switch": 1,
        "type": "heating",
    }


@pytest.mark.parametrize(
    ("index", "start", "step", "expected_time"),
    [
        (1, MIDNIGHT_DT, STEP, Time("00:30")),
        (47, MIDNIGHT_DT, STEP, Time("23:30")),
        (1, dt.datetime.strptime("01:00", TIME_FORMAT), STEP, Time("01:30")),
        (47, dt.datetime.strptime("01:00", TIME_FORMAT), STEP, Time("00:30")),
        (1, MIDNIGHT_DT, dt.timedelta(minutes=15), Time("00:15")),
    ],
)
def test_get_time(
    index: int, start: dt.datetime, step: dt.timedelta, expected_time: Time
) -> None:
    """Test returning time for a specific index."""
    assert get_time(index, start=start, step=step) == expected_time


@pytest.mark.parametrize(
    ("start", "end", "step", "expected_range"),
    [
        (
            Time("00:00"),
            Time("01:00"),
            STEP,
            [Time("00:00"), Time("00:30"), Time("01:00")],
        ),
        (
            Time("00:00"),
            Time("00:30"),
            dt.timedelta(minutes=15),
            [Time("00:00"), Time("00:15"), Time("00:30")],
        ),
        (
            Time("00:00"),
            Time("00:30"),
            dt.timedelta(minutes=20),
            [Time("00:00"), Time("00:20")],
        ),
        (
            Time("00:00"),
            Time("00:00"),
            dt.timedelta(hours=12),
            [Time("00:00"), Time("12:00")],
        ),
    ],
)
def test_get_time_range(
    start: Time, end: Time, step: dt.timedelta, expected_range: list[Time]
) -> None:
    """Test getting the time range."""
    assert get_time_range(start=start, end=end, step=step) == expected_range


def test_get_time_invalid() -> None:
    """Test getting the time range with invalid boundaries."""
    with pytest.raises(ValueError, match="Invalid time range"):
        get_time_range(start=Time("01:00"), end=Time("00:30"))


@pytest.fixture(name="intervals")
def fixture_intervals() -> list[Time]:
    """Return time intervals for a schedule day."""
    return (
        "00:00, 00:30, 01:00, 01:30, 02:00, 02:30, 03:00, 03:30, 04:00, 04:30, 05:00, "
        "05:30, 06:00, 06:30, 07:00, 07:30, 08:00, 08:30, 09:00, 09:30, 10:00, 10:30, "
        "11:00, 11:30, 12:00, 12:30, 13:00, 13:30, 14:00, 14:30, 15:00, 15:30, 16:00, "
        "16:30, 17:00, 17:30, 18:00, 18:30, 19:00, 19:30, 20:00, 20:30, 21:00, 21:30, "
        "22:00, 22:30, 23:00, 23:30"
    ).split(", ")


@pytest.fixture(name="schedule_day")
def fixture_schedule_day() -> ScheduleDay:
    """Return a schedule day object."""
    schedule_day = ScheduleDay.from_iterable([False for _ in range(48)])
    assert len(schedule_day) == 48
    return schedule_day


class TestScheduleDay:
    """Contains tests for ScheduleDay class."""

    @pytest.mark.parametrize(
        ("start", "end", "state", "error_pattern"),
        [
            (Time("00:00"), Time("01:00"), STATE_ON, None),
            (Time("01:00"), Time("10:00"), STATE_OFF, None),
            (Time("01:00"), Time("00:30"), RAISES, "Invalid time range"),
            (Time("00:foo"), Time("bar"), RAISES, "%H:%M"),
            (Time("05:00"), Time("18:00"), RAISES, "on, off"),
        ],
    )
    def test_set_state(
        self,
        start: Time,
        end: Time,
        state: Literal["on", "off", "raises"],
        error_pattern: str | None,
        schedule_day: ScheduleDay,
    ) -> None:
        """Test setting a state."""
        if state == RAISES:
            with pytest.raises((ValueError, TypeError), match=error_pattern):
                schedule_day.set_state(state, start, end)  # type: ignore[arg-type]
            return

        schedule_day.set_state(state, start, end)
        assert schedule_day[start] == state
        assert schedule_day[end] == state
        if state == STATE_ON:
            schedule_day[start] = STATE_OFF
            assert schedule_day.schedule[start] is False
        else:
            schedule_day[start] = STATE_ON
            assert schedule_day.schedule[start] is True

    def test_set_whole_day_schedule(
        self, schedule_day: ScheduleDay, intervals: list[Time]
    ) -> None:
        """Test setting schedule for a whole day."""
        schedule_day.set_on()
        assert schedule_day.schedule == dict.fromkeys(intervals, True)
        schedule_day.set_off()
        assert schedule_day.schedule == dict.fromkeys(intervals, False)

    def test_repr(self, schedule_day: ScheduleDay, intervals: list[Time]) -> None:
        """Test serializable representation of a schedule day."""
        schedule = dict.fromkeys(intervals, False)
        assert repr(schedule_day) == (f"ScheduleDay({schedule})")

    def test_iter(self, schedule_day: ScheduleDay) -> None:
        """Test setting schedule using iterable methods."""
        expected_times = list(schedule_day.schedule.keys())
        iterated_times = list(iter(schedule_day))
        assert iterated_times == expected_times

    def test_mapping(self, schedule_day: ScheduleDay) -> None:
        """Test mutable mapping methods."""
        assert schedule_day[MIDNIGHT] == STATE_OFF
        schedule_day[MIDNIGHT] = True
        assert schedule_day[MIDNIGHT] == STATE_ON
        del schedule_day[MIDNIGHT]
        assert len(schedule_day) == 47
        assert MIDNIGHT not in schedule_day


@pytest.fixture(name="schedule")
def fixture_schedule(schedule_day: ScheduleDay) -> Schedule:
    """Return a schedule object."""
    return Schedule(
        name="test",
        device=Mock(spec=PhysicalDevice, autospec=True),
        monday=ScheduleDay.from_iterable([True for _ in range(48)]),
        tuesday=schedule_day,
        wednesday=schedule_day,
        thursday=schedule_day,
        friday=schedule_day,
        saturday=schedule_day,
        sunday=schedule_day,
    )


class TestSchedule:
    """Contains tests for Schedule class."""

    @pytest.mark.parametrize(
        ("day", "time", "state"),
        [
            ("sunday", MIDNIGHT, STATE_OFF),
            ("monday", MIDNIGHT, STATE_ON),
            ("tuesday", MIDNIGHT, STATE_OFF),
            ("wednesday", MIDNIGHT, STATE_OFF),
            ("thursday", MIDNIGHT, STATE_OFF),
            ("friday", MIDNIGHT, STATE_OFF),
            ("saturday", MIDNIGHT, STATE_OFF),
        ],
    )
    def test_get(self, day: str, time: Time, state: State, schedule: Schedule) -> None:
        """Test getting a schedule day."""
        schedule_day = cast(ScheduleDay, getattr(schedule, day))
        assert schedule_day[time] == state

    def test_iter(self, schedule: Schedule) -> None:
        """Test a schedule."""
        expected_days = [
            schedule.sunday,
            schedule.monday,
            schedule.tuesday,
            schedule.wednesday,
            schedule.thursday,
            schedule.friday,
            schedule.saturday,
        ]
        iterated_days = list(iter(schedule))
        assert iterated_days == expected_days

    @patch("pyplumio.structures.schedules.Request.create")
    async def test_commit(self, mock_request_create, schedule: Schedule) -> None:
        """Test committing a schedule."""
        mock_device = Mock(spec=PhysicalDevice, autospec=True)
        mock_device.address = DeviceType.ECOMAX
        mock_device.data = {
            f"test_{ATTR_SCHEDULE_SWITCH}": 1,
            f"test_{ATTR_SCHEDULE_PARAMETER}": 2,
            ATTR_SCHEDULES: {"test": schedule},
        }
        mock_device.queue = Mock(spec=asyncio.Queue, autospec=True)
        schedule.device = mock_device
        await schedule.commit()
        mock_request_create.assert_awaited_once_with(
            FrameType.REQUEST_SET_SCHEDULE,
            recipient=mock_device.address,
            data={
                ATTR_TYPE: "test",
                ATTR_SWITCH: 1,
                ATTR_PARAMETER: 2,
                ATTR_SCHEDULE: schedule,
            },
        )



"""Contains structure decoders."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from pyplumio.frames import Frame


@dataclass
class StructureDataClass:
    """Represents a structure dataclass mixin."""

    __slots__ = ("frame",)

    frame: Frame


class Structure(ABC, StructureDataClass):
    """Represents a data structure."""

    __slots__ = ()

    @abstractmethod
    def encode(self, data: dict[str, Any]) -> bytearray:
        """Encode data to the bytearray message."""

    @abstractmethod
    def decode(
        self,
        message: bytearray,
        offset: int = 0,
        data: dict[str, Any] | None = None,
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""


class StructureDecoder(Structure, ABC):
    """Represents a structure that only handles decoding."""

    __slots__ = ()

    def encode(self, data: dict[str, Any]) -> bytearray:
        """Encode data to the bytearray message."""
        return bytearray()

    @abstractmethod
    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""


__all__ = ["Structure", "StructureDecoder"]



"""Contains tests for the object factory."""

from __future__ import annotations

from typing import Literal

import pytest

from pyplumio.frames import Frame, Request, Response
from pyplumio.frames.messages import RegulatorDataMessage
from pyplumio.frames.requests import StopMasterRequest
from pyplumio.frames.responses import UIDResponse
from pyplumio.helpers.factory import create_instance
from tests.conftest import RAISES


@pytest.mark.parametrize(
    ("path", "base", "expected", "error_pattern"),
    [
        ("frames.requests.StopMasterRequest", Request, StopMasterRequest, None),
        ("frames.responses.UIDResponse", Response, UIDResponse, None),
        ("frames.messages.RegulatorDataMessage", Frame, RegulatorDataMessage, None),
        ("frames.responses.UIDResponse", Request, RAISES, "Expected instance"),
        ("frames.requests.NonExistent", Request, RAISES, "no attribute"),
        ("frames.request.StopMasterRequest", Request, RAISES, "No module"),
    ],
)
async def test_create_instance(
    path: str,
    base: type,
    expected: type | Literal["raises"],
    error_pattern: str | None,
) -> None:
    """Test creating an instance of class."""
    if expected == RAISES:
        with pytest.raises(
            (TypeError, AttributeError, ModuleNotFoundError), match=error_pattern
        ):
            await create_instance(path, cls=base)
    else:
        assert isinstance(await create_instance(path, cls=base), expected)



"""Contains tests for the utility functions."""

import asyncio
from unittest.mock import Mock, patch

import pytest

from pyplumio import utils
from tests.conftest import RAISES


@pytest.mark.parametrize(
    ("input_value", "overrides", "expected"),
    [
        ("make_love_not_war", None, "MakeLoveNotWar"),
        ("make_love_not_war", {"love": "LOVE", "not": "not"}, "MakeLOVEnotWar"),
        ("miku_love_forever", None, "MikuLoveForever"),
        (
            "miku_love_forever",
            {"love": "LOVE", "forever": "Forever"},
            "MikuLOVEForever",
        ),
    ],
)
def test_to_camelcase(input_value, overrides, expected) -> None:
    """Test string to camelcase converter."""
    assert utils.to_camelcase(input_value, overrides=overrides) == expected


@pytest.mark.parametrize(
    ("dct", "defaults", "expected"),
    [
        ({"foo": "bar"}, {"baz": "foobar"}, {"foo": "bar", "baz": "foobar"}),
        ({}, {"baz": "foobar"}, {"baz": "foobar"}),
    ],
)
def test_ensure_dict(dct, defaults, expected) -> None:
    """Test ensure dictionary."""
    assert utils.ensure_dict(dct, defaults) == expected


@pytest.mark.parametrize(
    ("input_value", "divisor", "expected"),
    [
        (10.0, 0.2, True),
        (0.0, 1.0, True),
        (10.0, 3.0, False),
        (39.0, 13.0, True),
        (39.0, 7.0, False),
        (39.0, 1.0, True),
        (39.0, 39.0, True),
        (39.0, 0.0, RAISES),
    ],
)
def test_is_divisible(input_value, divisor, expected) -> None:
    """Test divisibility check."""
    if expected == RAISES:
        with pytest.raises(ValueError, match="Division by zero is not allowed."):
            utils.is_divisible(input_value, divisor)
    else:
        assert utils.is_divisible(input_value, divisor) == expected


@patch("asyncio.wait_for", side_effect=("test", asyncio.TimeoutError))
async def test_timeout(mock_wait_for) -> None:
    """Test a timeout decorator."""
    # Mock function to pass to the decorator.
    mock_func = Mock()
    mock_func.return_value = "test"

    # Call the decorator.
    timeout_decorator = utils.timeout(10)
    wrapper = timeout_decorator(mock_func)
    result = await wrapper("test_arg", kwarg="test_kwarg")
    assert result == "test"
    mock_wait_for.assert_awaited_once_with(mock_func.return_value, timeout=10)
    mock_func.assert_called_once_with("test_arg", kwarg="test_kwarg")

    # Test behavior when a timeout occurs.
    mock_func = Mock()
    mock_func.return_value = "test"
    decorator = utils.timeout(10)
    wrapper = decorator(mock_func)
    with pytest.raises(asyncio.TimeoutError):
        await wrapper("test_arg", kwarg="test_kwarg")



"""Contains device classes."""

from __future__ import annotations

from abc import ABC
import asyncio
from functools import cache
import logging
from typing import Any, ClassVar

from pyplumio.const import ATTR_FRAME_ERRORS, DeviceType, FrameType, State
from pyplumio.exceptions import RequestError, UnknownDeviceError
from pyplumio.filters import on_change
from pyplumio.frames import Frame, Request, is_known_frame_type
from pyplumio.helpers.event_manager import EventManager, event_listener
from pyplumio.helpers.factory import create_instance
from pyplumio.parameters import NumericType, Parameter
from pyplumio.structures.network_info import NetworkInfo
from pyplumio.utils import to_camelcase

_LOGGER = logging.getLogger(__name__)


@cache
def is_known_device_type(device_type: int) -> bool:
    """Check if device type is known."""
    try:
        DeviceType(device_type)
        return True
    except ValueError:
        return False


@cache
def get_device_handler(device_type: int) -> str:
    """Return module name and class name for a given device type."""
    if not is_known_device_type(device_type):
        raise UnknownDeviceError(f"Unknown device type ({device_type})")

    type_name = to_camelcase(
        DeviceType(device_type).name,
        overrides={
            "ecomax": "EcoMAX",
            "ecoster": "EcoSTER",
            "econet": "EcoNET",
        },
    )
    return f"devices.{type_name.lower()}.{type_name}"


class Device(ABC, EventManager):
    """Represents a device."""

    __slots__ = ("queue",)

    queue: asyncio.Queue[Frame]

    def __init__(self, queue: asyncio.Queue[Frame]) -> None:
        """Initialize a new device."""
        super().__init__()
        self.queue = queue

    async def set(
        self,
        name: str,
        value: NumericType | State | bool,
        retries: int = 0,
        timeout: float | None = None,
    ) -> bool:
        """Set a parameter value.

        :param name: Name of the parameter
        :type name: str
        :param value: New value for the parameter
        :type value: int | float | bool | Literal["on", "off"]
        :param retries: Try setting parameter for this amount of
            times, defaults to 0 (disabled)
        :type retries: int, optional
        :param timeout: Wait this amount of seconds for confirmation,
            defaults to `None`
        :type timeout: float, optional
        :return: `True` if parameter was successfully set, `False`
            otherwise.
        :rtype: bool
        :raise asyncio.TimeoutError: when waiting past specified timeout
        :raise ValueError: when a new value is outside of allowed range
        :raise TypeError: when found data is not valid parameter
        """
        parameter = await self.get(name, timeout)
        if not isinstance(parameter, Parameter):
            raise TypeError(f"The parameter '{name}' is not valid or does not exist.")

        return await parameter.set(value, retries=retries)

    def set_nowait(
        self,
        name: str,
        value: NumericType | State | bool,
        retries: int = 0,
        timeout: float | None = None,
    ) -> None:
        """Set a parameter value without waiting for the result.

        :param name: Name of the parameter
        :type name: str
        :param value: New value for the parameter
        :type value: int | float | bool | Literal["on", "off"]
        :param retries: Try setting parameter for this amount of
            times, defaults to 0 (disabled)
        :type retries: int, optional
        :param timeout: Wait this amount of seconds for confirmation.
            As this method operates in the background without waiting,
            this value is used to determine failure when
            retrying and doesn't block, defaults to `None`
        :type timeout: float, optional
        """
        self.create_task(self.set(name, value, retries, timeout))

    async def shutdown(self) -> None:
        """Cancel device tasks."""
        self.cancel_tasks()
        await self.wait_until_done()


class PhysicalDevice(Device, ABC):
    """Represents a physical device.

    Physical device have network address and can have multiple
    virtual devices associated with them via parent property.
    """

    __slots__ = ("address", "_network", "_frame_versions")

    address: ClassVar[int]

    _network: NetworkInfo
    _frame_versions: dict[int, int]

    def __init__(self, queue: asyncio.Queue[Frame], network: NetworkInfo) -> None:
        """Initialize a new physical device."""
        super().__init__(queue)
        self._network = network
        self._frame_versions = {}

    @event_listener(filter=on_change)
    async def on_event_frame_versions(self, versions: dict[int, int]) -> None:
        """Check frame versions and update outdated frames."""
        _LOGGER.debug("Received frame version table")
        for frame_type, version in versions.items():
            if (
                is_known_frame_type(frame_type)
                and self.supports_frame_type(frame_type)
                and not self.has_frame_version(frame_type, version)
            ):
                request_frame_type = (
                    FrameType.REQUEST_ECOMAX_PARAMETERS
                    if frame_type == FrameType.REQUEST_ECOMAX_PARAMETER_CHANGES
                    else frame_type
                )
                await self._request_frame_version(request_frame_type, version)
                self._frame_versions[frame_type] = version

    async def _request_frame_version(
        self, frame_type: FrameType | int, version: int
    ) -> None:
        """Request frame version from the device."""
        _LOGGER.debug("Updating frame %s to version %i", repr(frame_type), version)
        request = await Request.create(frame_type, recipient=self.address)
        self.queue.put_nowait(request)

    def has_frame_version(self, frame_type: FrameType | int, version: int) -> bool:
        """Return True if frame data is up to date, False otherwise."""
        return (
            frame_type in self._frame_versions
            and self._frame_versions[frame_type] == version
        )

    def supports_frame_type(self, frame_type: int) -> bool:
        """Check if frame type is supported by the device."""
        return frame_type not in self.data.get(ATTR_FRAME_ERRORS, [])

    def handle_frame(self, frame: Frame) -> None:
        """Handle frame received from the device."""
        frame.assign_to(self)
        if frame.data is not None:
            for name, value in frame.data.items():
                self.dispatch_nowait(name, value)

    async def request(
        self, name: str, frame_type: FrameType, retries: int = 3, timeout: float = 3.0
    ) -> Any:
        """Send request and wait for a value to become available.

        If value is not available before timeout, retry request.
        """
        _LOGGER.info("Requesting '%s' with %s", name, repr(frame_type))
        request = await Request.create(frame_type, recipient=self.address)
        while retries > 0:
            try:
                self.queue.put_nowait(request)
                return await self.get(name, timeout=timeout)
            except asyncio.TimeoutError:
                retries -= 1

        raise RequestError(
            f"Failed to request '{name}' with frame type '{frame_type}' after "
            f"{retries} retries.",
            frame_type=frame_type,
        )

    @classmethod
    async def create(cls, device_type: int, **kwargs: Any) -> PhysicalDevice:
        """Create a physical device handler object."""
        return await create_instance(get_device_handler(device_type), cls=cls, **kwargs)


class VirtualDevice(Device, ABC):
    """Represents a virtual device associated with physical device."""

    __slots__ = ("parent", "index")

    parent: PhysicalDevice
    index: int

    def __init__(
        self, queue: asyncio.Queue[Frame], parent: PhysicalDevice, index: int = 0
    ) -> None:
        """Initialize a new sub-device."""
        super().__init__(queue)
        self.parent = parent
        self.index = index


__all__ = [
    "Device",
    "PhysicalDevice",
    "VirtualDevice",
    "is_known_device_type",
    "get_device_handler",
]



"""Contains mixer parameter descriptors."""

from __future__ import annotations

from dataclasses import dataclass
from functools import cache
from typing import TYPE_CHECKING

from dataslots import dataslots

from pyplumio.const import (
    ATTR_DEVICE_INDEX,
    ATTR_INDEX,
    ATTR_VALUE,
    FrameType,
    ProductType,
    UnitOfMeasurement,
)
from pyplumio.frames import Request
from pyplumio.parameters import (
    OffsetNumber,
    OffsetNumberDescription,
    Parameter,
    ParameterDescription,
    Switch,
    SwitchDescription,
)
from pyplumio.structures.product_info import ProductInfo

if TYPE_CHECKING:
    from pyplumio.devices.mixer import Mixer


@dataclass
class MixerParameterDescription(ParameterDescription):
    """Represents a mixer parameter description."""

    __slots__ = ()


class MixerParameter(Parameter):
    """Represent a mixer parameter."""

    __slots__ = ()

    device: Mixer
    description: MixerParameterDescription

    async def create_request(self) -> Request:
        """Create a request to change the parameter."""
        return await Request.create(
            FrameType.REQUEST_SET_MIXER_PARAMETER,
            recipient=self.device.parent.address,
            data={
                ATTR_INDEX: self._index,
                ATTR_VALUE: self.values.value,
                ATTR_DEVICE_INDEX: self.device.index,
            },
        )


@dataclass
class MixerNumberDescription(MixerParameterDescription, OffsetNumberDescription):
    """Represent a mixer number description."""

    __slots__ = ()


class MixerNumber(MixerParameter, OffsetNumber):
    """Represents a mixer number."""

    __slots__ = ()

    description: MixerNumberDescription


@dataslots
@dataclass
class MixerSwitchDescription(MixerParameterDescription, SwitchDescription):
    """Represents a mixer switch description."""


class MixerSwitch(MixerParameter, Switch):
    """Represents a mixer switch."""

    __slots__ = ()

    description: MixerSwitchDescription


PARAMETER_TYPES: dict[ProductType, list[MixerParameterDescription]] = {
    ProductType.ECOMAX_P: [
        MixerNumberDescription(
            name="mixer_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="min_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="max_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="thermostat_decrease_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerSwitchDescription(
            name="weather_control",
        ),
        MixerNumberDescription(
            name="heating_curve",
            step=0.1,
        ),
        MixerNumberDescription(
            name="heating_curve_shift",
            offset=20,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="weather_factor",
        ),
        MixerNumberDescription(
            name="work_mode",
        ),
        MixerNumberDescription(
            name="mixer_input_dead_zone",
            step=0.1,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerSwitchDescription(
            name="thermostat_operation",
        ),
        MixerNumberDescription(
            name="thermostat_mode",
        ),
        MixerSwitchDescription(
            name="disable_pump_on_thermostat",
        ),
        MixerSwitchDescription(
            name="summer_work",
        ),
    ],
    ProductType.ECOMAX_I: [
        MixerNumberDescription(
            name="work_mode",
        ),
        MixerNumberDescription(
            name="circuit_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="day_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="night_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="min_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="max_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerSwitchDescription(
            name="summer_work",
        ),
        MixerSwitchDescription(
            name="weather_control",
        ),
        MixerNumberDescription(
            name="enable_circuit",
        ),
        MixerNumberDescription(
            name="constant_water_preset_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="thermostat_decrease_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="thermostat_correction",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerSwitchDescription(
            name="thermostat_pump_lock",
        ),
        MixerNumberDescription(
            name="valve_opening_time",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        MixerNumberDescription(
            name="mixer_input_dead_zone",
            step=0.1,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="proportional_range",
        ),
        MixerNumberDescription(
            name="integration_time_constant",
        ),
        MixerNumberDescription(
            name="heating_curve",
            step=0.1,
        ),
        MixerNumberDescription(
            name="heating_curve_shift",
            offset=20,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="thermostat_mode",
        ),
        MixerNumberDescription(
            name="decreasing_constant_water_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
    ],
}


@cache
def get_mixer_parameter_types(
    product_info: ProductInfo,
) -> list[MixerParameterDescription]:
    """Return cached mixer parameter types for specific product."""
    return PARAMETER_TYPES[product_info.type]


__all__ = [
    "get_mixer_parameter_types",
    "MixerNumber",
    "MixerNumberDescription",
    "MixerParameter",
    "MixerParameterDescription",
    "MixerSwitch",
    "MixerSwitchDescription",
    "PARAMETER_TYPES",
]



"""Contains a factory helper."""

from __future__ import annotations

import asyncio
import importlib
import logging
from types import ModuleType
from typing import Any, TypeVar

from pyplumio.helpers.async_cache import acache

_LOGGER = logging.getLogger(__name__)

T = TypeVar("T")


@acache
async def import_module(name: str) -> ModuleType:
    """Import module by name."""
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, importlib.import_module, f"pyplumio.{name}")


async def create_instance(class_path: str, /, cls: type[T], **kwargs: Any) -> T:
    """Return a class instance from the class path."""
    module_name, class_name = class_path.rsplit(".", 1)
    module = await import_module(module_name)
    instance = getattr(module, class_name)(**kwargs)
    if not isinstance(instance, cls):
        raise TypeError(
            f"Expected instance of '{cls.__name__}', but got "
            f"'{type(instance).__name__}' from '{class_name}'"
        )

    return instance


__all__ = ["create_instance"]



"""Contains tests for the mixer parameter descriptors."""

import pytest

from pyplumio.const import ATTR_DEVICE_INDEX, ATTR_INDEX, ATTR_VALUE, ProductType
from pyplumio.devices.ecomax import EcoMAX
from pyplumio.devices.mixer import Mixer
from pyplumio.frames.requests import SetMixerParameterRequest
from pyplumio.parameters import ParameterValues
from pyplumio.parameters.mixer import (
    PARAMETER_TYPES as MIXER_PARAMETER_TYPE,
    MixerNumber,
    MixerNumberDescription,
    MixerSwitch,
    MixerSwitchDescription,
    get_mixer_parameter_types,
)


@pytest.fixture(name="mixer")
def fixture_mixer(ecomax: EcoMAX) -> Mixer:
    """Return an mixer object."""
    return Mixer(ecomax.queue, parent=ecomax)


async def test_mixer_parameter_create_request(mixer: Mixer) -> None:
    """Test create_request method for mixer parameter."""
    # Test with number.
    mixer_number = MixerNumber(
        device=mixer,
        description=MixerNumberDescription(name="test_number"),
        values=ParameterValues(value=10, min_value=0, max_value=20),
    )
    mixer_number_request = await mixer_number.create_request()
    assert isinstance(mixer_number_request, SetMixerParameterRequest)
    assert mixer_number_request.data == {
        ATTR_INDEX: 0,
        ATTR_VALUE: 10,
        ATTR_DEVICE_INDEX: 0,
    }

    # Test with switch.
    mixer_switch = MixerSwitch(
        device=mixer,
        description=MixerSwitchDescription(name="test_switch"),
        values=ParameterValues(value=0, min_value=0, max_value=1),
    )
    mixer_switch_request = await mixer_switch.create_request()
    assert isinstance(mixer_switch_request, SetMixerParameterRequest)
    assert mixer_switch_request.data == {
        ATTR_INDEX: 0,
        ATTR_VALUE: 0,
        ATTR_DEVICE_INDEX: 0,
    }


def test_get_mixer_parameter_types(mixer: Mixer) -> None:
    """Test ecoMAX parameter types getter."""
    parameter_types = get_mixer_parameter_types(mixer.parent.product)
    assert len(MIXER_PARAMETER_TYPE[ProductType.ECOMAX_P]) == 14
    assert MIXER_PARAMETER_TYPE[ProductType.ECOMAX_P] == parameter_types



"""Contains tests for ecoMAX parameters structure decoder."""

import pytest

from pyplumio.frames.responses import EcomaxParametersResponse
from pyplumio.parameters import ParameterValues
from pyplumio.structures.ecomax_parameters import (
    ATTR_ECOMAX_PARAMETERS,
    EcomaxParametersStructure,
)


@pytest.fixture(name="ecomax_parameters_structure")
def fixture_ecomax_parameters_structure() -> EcomaxParametersStructure:
    """Fixture for EcomaxParametersResponse."""
    return EcomaxParametersStructure(frame=EcomaxParametersResponse())


class TestEcomaxParametersStructure:
    """Test the EcomaxParameters structure decoder."""

    def test_decode(
        self, ecomax_parameters_structure: EcomaxParametersStructure
    ) -> None:
        """Test decoding of ecoMAX parameters structure."""
        message = bytearray([0x01, 0x00, 0x01, 0x02, 0x01, 0x03])
        data, offset = ecomax_parameters_structure.decode(message)
        assert data == {
            ATTR_ECOMAX_PARAMETERS: [
                (0, ParameterValues(value=2, min_value=1, max_value=3))
            ]
        }
        assert offset == 6



"""Contains tests for the custom parameter descriptors."""

from unittest.mock import patch

import pytest

from pyplumio.const import ProductType
from pyplumio.parameters.custom import (
    CustomParameter,
    CustomParameters,
    Signature,
    inject_custom_parameters,
)
from pyplumio.parameters.ecomax import (
    PARAMETER_TYPES as ECOMAX_PARAMETER_TYPES,
    EcomaxNumberDescription,
)
from pyplumio.structures.product_info import ProductInfo


class DummyEcoMAX(CustomParameters):
    """Replacements for ecoMAX 860D3-HB."""

    __slots__ = ()

    signature = Signature(model="ecoMAX Dummy", id=1)

    replacements = (
        CustomParameter(
            original="summer_mode_disable_temp",
            replacement=EcomaxNumberDescription(name="__test_replacement_1"),
        ),
    )


@pytest.mark.parametrize(
    ("product_info", "module_name", "expected_result"),
    [
        (
            ProductInfo(
                type=ProductType.ECOMAX_P,
                id=1,
                uid="*TEST*",
                logo=1,
                image=1,
                model="ecoMAX Dummy",
            ),
            "ecomax_dummy.EcoMAXDUMMY",
            True,
        ),
        (
            ProductInfo(
                type=ProductType.ECOMAX_P,
                id=2,
                uid="*TEST*",
                logo=2,
                image=1,
                model="ecoMAX Dummy",
            ),
            "ecomax_dummy.EcoMAXDUMMY",
            False,
        ),
        (
            ProductInfo(
                type=ProductType.ECOMAX_P,
                id=1,
                uid="*TEST*",
                logo=1,
                image=1,
                model="ecoMAX Dummy2",
            ),
            "ecomax_dummy2.EcoMAXDUMMY2",
            False,
        ),
    ],
)
@patch("pyplumio.parameters.custom.create_instance", return_value=DummyEcoMAX())
async def test_inject_custom_parameters(
    moke_create_instance,
    product_info: ProductInfo,
    expected_result: bool,
    module_name: str,
) -> None:
    """Test custom parameters injection."""
    parameters = await inject_custom_parameters(
        product_info, parameter_types=ECOMAX_PARAMETER_TYPES[ProductType.ECOMAX_P]
    )
    replacement_found = any(
        parameter
        for parameter in parameters
        if parameter.name == "__test_replacement_1"
    )
    assert replacement_found is expected_result, (
        f"Replacements {'not' if expected_result else ''} found: {parameters}"
    )
    moke_create_instance.assert_called_once_with(
        f"parameters.custom.{module_name}", cls=CustomParameters
    )



"""Contains tests for fuel consumption structure decoder."""

import pytest

from pyplumio.frames.messages import SensorDataMessage
from pyplumio.structures.fuel_consumption import (
    ATTR_FUEL_CONSUMPTION,
    FuelConsumptionStructure,
)


@pytest.fixture(name="fuel_consumption_structure")
def fixture_fuel_consumption_structure() -> FuelConsumptionStructure:
    """Fixture for FuelConsumptionStructure."""
    return FuelConsumptionStructure(frame=SensorDataMessage())


class TestFuelConsumptionStructure:
    """Test the FuelConsumption structure decoder."""

    def test_decode(self, fuel_consumption_structure: FuelConsumptionStructure) -> None:
        """Test decoding of fuel consumption structure."""
        message = bytearray([0x00, 0x00, 0x48, 0x42])
        data, offset = fuel_consumption_structure.decode(message)
        assert data == {ATTR_FUEL_CONSUMPTION: 50.0}
        assert offset == 4

    def test_decode_with_nan(
        self, fuel_consumption_structure: FuelConsumptionStructure
    ) -> None:
        """Test decoding of fuel consumption structure with NaN value."""
        message = bytearray([0x00, 0x00, 0xC0, 0x7F])  # Represents NaN
        data, offset = fuel_consumption_structure.decode(message)
        assert data == {}
        assert offset == 4



"""Contains a device parameter class."""

from __future__ import annotations

from abc import ABC, abstractmethod
import asyncio
from contextlib import suppress
from dataclasses import asdict, dataclass
import logging
from typing import TYPE_CHECKING, Any, Literal, TypeVar, Union, get_args

from dataslots import dataslots
from typing_extensions import TypeAlias

from pyplumio.const import BYTE_UNDEFINED, STATE_OFF, STATE_ON, State, UnitOfMeasurement
from pyplumio.frames import Request
from pyplumio.utils import is_divisible

if TYPE_CHECKING:
    from pyplumio.devices import Device

_LOGGER = logging.getLogger(__name__)

ParameterT = TypeVar("ParameterT", bound="Parameter")


def unpack_parameter(
    data: bytearray, offset: int = 0, size: int = 1
) -> ParameterValues | None:
    """Unpack a device parameter."""
    if not is_valid_parameter(data[offset : offset + size * 3]):
        return None

    value = data[offset : offset + size]
    min_value = data[offset + size : offset + 2 * size]
    max_value = data[offset + 2 * size : offset + 3 * size]

    return ParameterValues(
        value=int.from_bytes(value, byteorder="little"),
        min_value=int.from_bytes(min_value, byteorder="little"),
        max_value=int.from_bytes(max_value, byteorder="little"),
    )


def is_valid_parameter(data: bytearray) -> bool:
    """Check if parameter contains any bytes besides 0xFF."""
    return any(x for x in data if x != BYTE_UNDEFINED)


@dataclass
class ParameterValues:
    """Represents a parameter values."""

    __slots__ = ("value", "min_value", "max_value")

    value: int
    min_value: int
    max_value: int


@dataslots
@dataclass
class ParameterDescription:
    """Represents a parameter description."""

    name: str
    optimistic: bool = False


NumericType: TypeAlias = Union[int, float]


class Parameter(ABC):
    """Represents a base parameter."""

    __slots__ = (
        "device",
        "description",
        "_update_done",
        "_update_pending",
        "_index",
        "_values",
    )

    device: Device
    description: ParameterDescription
    _update_done: asyncio.Event
    _update_pending: asyncio.Event
    _index: int
    _values: ParameterValues

    def __init__(
        self,
        device: Device,
        description: ParameterDescription,
        values: ParameterValues | None = None,
        index: int = 0,
    ):
        """Initialize a new parameter."""
        self.device = device
        self.description = description
        self._index = index
        self._index = index
        self._update_done = asyncio.Event()
        self._update_pending = asyncio.Event()
        self._values = values if values else ParameterValues(0, 0, 0)

    def __repr__(self) -> str:
        """Return a serializable string representation."""
        return (
            f"{self.__class__.__name__}("
            f"device={self.device}, "
            f"description={self.description}, "
            f"values={self.values}, "
            f"index={self._index})"
        )

    def __hash__(self) -> int:
        """Return a hash of the parameter based on its values."""
        return hash(frozenset(asdict(self.values).items()))

    def _call_relational_method(self, method_to_call: str, other: Any) -> Any:
        """Call a specified relational method."""
        if isinstance(other, Parameter):
            other = other.values

        if isinstance(other, ParameterValues):
            handler = getattr(self.values.value, method_to_call)
            return handler(other.value)

        if isinstance(other, (int, float, bool)) or other in get_args(State):
            handler = getattr(self.values.value, method_to_call)
            return handler(self._pack_value(other))
        else:
            return NotImplemented

    def __int__(self) -> int:
        """Return an integer representation of parameter's value."""
        return self.values.value

    def __add__(self, other: Any) -> Any:
        """Return result of addition."""
        return self._call_relational_method("__add__", other)

    def __sub__(self, other: Any) -> Any:
        """Return result of the subtraction."""
        return self._call_relational_method("__sub__", other)

    def __truediv__(self, other: Any) -> Any:
        """Return result of true division."""
        return self._call_relational_method("__truediv__", other)

    def __floordiv__(self, other: Any) -> Any:
        """Return result of floored division."""
        return self._call_relational_method("__floordiv__", other)

    def __mul__(self, other: Any) -> Any:
        """Return result of the multiplication."""
        return self._call_relational_method("__mul__", other)

    def __eq__(self, other: Any) -> Any:
        """Compare if parameter value is equal to other."""
        return self._call_relational_method("__eq__", other)

    def __ge__(self, other: Any) -> Any:
        """Compare if parameter value is greater or equal to other."""
        return self._call_relational_method("__ge__", other)

    def __gt__(self, other: Any) -> Any:
        """Compare if parameter value is greater than other."""
        return self._call_relational_method("__gt__", other)

    def __le__(self, other: Any) -> Any:
        """Compare if parameter value is less or equal to other."""
        return self._call_relational_method("__le__", other)

    def __lt__(self, other: Any) -> Any:
        """Compare if parameter value is less that other."""
        return self._call_relational_method("__lt__", other)

    def __copy__(self) -> Parameter:
        """Create a copy of parameter."""
        values = type(self.values)(
            self.values.value, self.values.min_value, self.values.max_value
        )
        return type(self)(self.device, self.description, values)

    async def set(self, value: Any, retries: int = 0, timeout: float = 5.0) -> bool:
        """Set a parameter value."""
        self.validate(value)
        return await self._attempt_update(self._pack_value(value), retries, timeout)

    def set_nowait(self, value: Any, retries: int = 0, timeout: float = 5.0) -> None:
        """Set a parameter value without waiting."""
        self.validate(value)
        self.device.create_task(
            self._attempt_update(self._pack_value(value), retries, timeout)
        )

    async def _attempt_update(self, value: int, retries: int, timeout: float) -> bool:
        """Attempt to update a parameter value on the remote device."""
        _LOGGER.info(
            "Attempting to update '%s' parameter to %d", self.description.name, value
        )
        if value == self.values.value:
            # Value is unchanged
            return True

        self._values.value = value
        request = await self.create_request()
        if self.description.optimistic:
            await self.device.queue.put(request)
            return True

        self.update_done.clear()
        self.update_pending.set()
        if retries > 0:
            return await self._attempt_update_with_retries(
                request, retries=retries, timeout=timeout
            )

        return await self._send_update_request(request, timeout=timeout)

    async def _attempt_update_with_retries(
        self, request: Request, retries: int, timeout: float
    ) -> bool:
        """Send update request and retry until success."""
        for _ in range(retries):
            if await self._send_update_request(request, timeout=timeout):
                return True

        _LOGGER.warning(
            "Unable to confirm update of '%s' parameter after %d retries",
            self.description.name,
            retries,
        )
        return False

    async def _send_update_request(self, request: Request, timeout: float) -> bool:
        """Send update request to the remote and confirm the result."""
        await self.device.queue.put(request)
        with suppress(asyncio.TimeoutError):
            # Wait for the update to be done
            await asyncio.wait_for(self.update_done.wait(), timeout=timeout)

        return self.update_done.is_set()

    def update(self, values: ParameterValues) -> None:
        """Update the parameter values."""
        self.update_done.set()
        self.update_pending.clear()
        self._values = values

    @property
    def update_done(self) -> asyncio.Event:
        """Check if parameter is updated on the device."""
        return self._update_done

    @property
    def update_pending(self) -> asyncio.Event:
        """Check if parameter is updated on the device."""
        return self._update_pending

    @property
    def values(self) -> ParameterValues:
        """Return the parameter values."""
        return self._values

    @classmethod
    def create_or_update(
        cls: type[ParameterT],
        device: Device,
        description: ParameterDescription,
        values: ParameterValues,
        **kwargs: Any,
    ) -> ParameterT:
        """Create new parameter or update parameter values."""
        parameter: ParameterT | None = device.get_nowait(description.name, None)
        if parameter and isinstance(parameter, cls):
            parameter.update(values)
        else:
            parameter = cls(
                device=device, description=description, values=values, **kwargs
            )

        return parameter

    @abstractmethod
    def _pack_value(self, value: Any) -> int:
        """Pack the parameter value."""

    @abstractmethod
    def _unpack_value(self, value: int) -> Any:
        """Unpack the parameter value."""

    @abstractmethod
    def validate(self, value: Any) -> bool:
        """Validate a parameter value."""

    @property
    @abstractmethod
    def value(self) -> NumericType | State | bool:
        """Return the value."""

    @property
    @abstractmethod
    def min_value(self) -> NumericType | State | bool:
        """Return the minimum allowed value."""

    @property
    @abstractmethod
    def max_value(self) -> NumericType | State | bool:
        """Return the maximum allowed value."""

    @abstractmethod
    async def create_request(self) -> Request:
        """Create a request to change the parameter."""


@dataslots
@dataclass
class NumberDescription(ParameterDescription):
    """Represents a parameter description."""

    step: float = 1.0
    precision: int = 6
    unit_of_measurement: UnitOfMeasurement | Literal["%"] | None = None


class Number(Parameter):
    """Represents a number."""

    __slots__ = ()

    description: NumberDescription

    def _pack_value(self, value: NumericType) -> int:
        """Pack the parameter value."""
        return int(round(value / self.description.step))

    def _unpack_value(self, value: int) -> NumericType:
        """Unpack the parameter value."""
        return round(value * self.description.step, self.description.precision)

    def validate(self, value: Any) -> bool:
        """Validate a parameter value."""
        if value < self.min_value or value > self.max_value:
            raise ValueError(
                f"Invalid value: {value}. The value must be between {self.min_value} "
                f"and {self.max_value}."
            )

        if not is_divisible(value, self.description.step, self.description.precision):
            raise ValueError(
                f"Invalid value: {value}. The value must be adjusted in increments of "
                f"{self.description.step}."
            )

        return True

    async def set(
        self, value: NumericType, retries: int = 0, timeout: float = 5.0
    ) -> bool:
        """Set a parameter value."""
        return await super().set(value, retries=retries, timeout=timeout)

    def set_nowait(
        self, value: NumericType, retries: int = 0, timeout: float = 5.0
    ) -> None:
        """Set a parameter value without waiting."""
        super().set_nowait(value, retries=retries, timeout=timeout)

    async def create_request(self) -> Request:
        """Create a request to change the number."""
        return Request()

    @property
    def value(self) -> NumericType:
        """Return the value."""
        return self._unpack_value(self.values.value)

    @property
    def min_value(self) -> NumericType:
        """Return the minimum allowed value."""
        return self._unpack_value(self.values.min_value)

    @property
    def max_value(self) -> NumericType:
        """Return the maximum allowed value."""
        return self._unpack_value(self.values.max_value)

    @property
    def unit_of_measurement(self) -> UnitOfMeasurement | Literal["%"] | None:
        """Return the unit of measurement."""
        return self.description.unit_of_measurement


@dataslots
@dataclass
class OffsetNumberDescription(NumberDescription):
    """Represents a parameter description."""

    offset: int = 0


class OffsetNumber(Number):
    """Represents a number with offset."""

    __slots__ = ()

    description: OffsetNumberDescription

    def _pack_value(self, value: NumericType) -> int:
        """Pack the parameter value."""
        return super()._pack_value(value + self.description.offset)

    def _unpack_value(self, value: int) -> NumericType:
        """Unpack the parameter value."""
        return super()._unpack_value(value - self.description.offset)


@dataslots
@dataclass
class SwitchDescription(ParameterDescription):
    """Represents a switch description."""


class Switch(Parameter):
    """Represents a switch."""

    __slots__ = ()

    description: SwitchDescription

    def _pack_value(self, value: State | bool) -> int:
        """Pack the parameter value."""
        if value in get_args(State):
            return 1 if value == STATE_ON else 0

        return int(value)

    def _unpack_value(self, value: int) -> State:
        """Unpack the parameter value."""
        return STATE_ON if value == 1 else STATE_OFF

    def validate(self, value: Any) -> bool:
        """Validate a parameter value."""
        if not isinstance(value, bool) and value not in get_args(State):
            raise ValueError(
                f"Invalid value: {value}. The value must be either 'on', 'off' or "
                f"boolean."
            )

        return True

    async def set(
        self, value: State | bool, retries: int = 0, timeout: float = 5.0
    ) -> bool:
        """Set a parameter value."""
        return await super().set(value, retries=retries, timeout=timeout)

    def set_nowait(
        self, value: State | bool, retries: int = 0, timeout: float = 5.0
    ) -> None:
        """Set a switch value without waiting."""
        super().set_nowait(value, retries=retries, timeout=timeout)

    async def turn_on(self) -> bool:
        """Set a switch value to 'on'.

        :return: `True` if parameter was successfully turned on, `False`
            otherwise.
        :rtype: bool
        """
        return await self.set(STATE_ON)

    async def turn_off(self) -> bool:
        """Set a switch value to 'off'.

        :return: `True` if parameter was successfully turned off, `False`
            otherwise.
        :rtype: bool
        """
        return await self.set(STATE_OFF)

    def turn_on_nowait(self) -> None:
        """Set a switch value to 'on' without waiting."""
        self.set_nowait(STATE_ON)

    def turn_off_nowait(self) -> None:
        """Set a switch value to 'off' without waiting."""
        self.set_nowait(STATE_OFF)

    async def create_request(self) -> Request:
        """Create a request to change the switch."""
        return Request()

    @property
    def value(self) -> State:
        """Return the value."""
        return self._unpack_value(self.values.value)

    @property
    def min_value(self) -> Literal["off"]:
        """Return the minimum allowed value."""
        return STATE_OFF

    @property
    def max_value(self) -> Literal["on"]:
        """Return the maximum allowed value."""
        return STATE_ON


__all__ = [
    "Number",
    "NumberDescription",
    "NumericType",
    "OffsetNumber",
    "OffsetNumberDescription",
    "Parameter",
    "ParameterDescription",
    "ParameterValues",
    "State",
    "Switch",
    "SwitchDescription",
]



"""Contains patch for ecoMAX 860D3-HB."""

from pyplumio.const import UnitOfMeasurement
from pyplumio.parameters.custom import CustomParameter, CustomParameters, Signature
from pyplumio.parameters.ecomax import EcomaxNumberDescription, EcomaxSwitchDescription


class EcoMAX860D3HB(CustomParameters):
    """Replacements for ecoMAX 860D3-HB."""

    signature = Signature(model="ecoMAX 860D3-HB", id=48)

    replacements = (
        # Summer mode
        CustomParameter(
            original="summer_mode_disable_temp",
            replacement=EcomaxNumberDescription(name="__unknown_parameter_1"),
        ),
        CustomParameter(
            original="water_heater_target_temp",
            replacement=EcomaxNumberDescription(name="summer_mode"),
        ),
        CustomParameter(
            original="min_water_heater_target_temp",
            replacement=EcomaxNumberDescription(
                name="summer_mode_enable_temp",
                unit_of_measurement=UnitOfMeasurement.CELSIUS,
            ),
        ),
        CustomParameter(
            original="max_water_heater_target_temp",
            replacement=EcomaxNumberDescription(
                name="summer_mode_disable_temp",
                unit_of_measurement=UnitOfMeasurement.CELSIUS,
            ),
        ),
        # Water heater
        CustomParameter(
            original="disable_pump_on_thermostat",
            replacement=EcomaxNumberDescription(
                name="water_heater_target_temp",
                unit_of_measurement=UnitOfMeasurement.CELSIUS,
            ),
        ),
        CustomParameter(
            original="boiler_alert_temp",
            replacement=EcomaxNumberDescription(
                name="min_water_heater_target_temp",
                unit_of_measurement=UnitOfMeasurement.CELSIUS,
            ),
        ),
        CustomParameter(
            original="max_feeder_temp",
            replacement=EcomaxNumberDescription(
                name="max_water_heater_target_temp",
                unit_of_measurement=UnitOfMeasurement.CELSIUS,
            ),
        ),
        CustomParameter(
            original="water_heater_work_mode",
            replacement=EcomaxNumberDescription(name="water_heater_feeding_extension"),
        ),
        CustomParameter(
            original="external_boiler_temp",
            replacement=EcomaxNumberDescription(name="water_heater_work_mode"),
        ),
        CustomParameter(
            original="alert_notify",
            replacement=EcomaxNumberDescription(
                name="water_heater_hysteresis",
                unit_of_measurement=UnitOfMeasurement.CELSIUS,
            ),
        ),
        CustomParameter(
            original="pump_hysteresis",
            replacement=EcomaxSwitchDescription(name="water_heater_disinfection"),
        ),
    )



"""Contains tests for mixer parameters structure decoder."""

import pytest

from pyplumio.frames.responses import MixerParametersResponse
from pyplumio.parameters import ParameterValues
from pyplumio.structures.mixer_parameters import (
    ATTR_MIXER_PARAMETERS,
    MixerParametersStructure,
)


@pytest.fixture(name="mixer_parameters_structure")
def fixture_mixer_parameters_structure() -> MixerParametersStructure:
    """Fixture for MixerParametersStructure."""
    return MixerParametersStructure(frame=MixerParametersResponse())


class TestMixerParametersStructure:
    """Test the MixerParameters structure decoder."""

    def test_decode(self, mixer_parameters_structure: MixerParametersStructure) -> None:
        """Test decoding of mixer parameters structure."""
        message = bytearray([0x01, 0x00, 0x01, 0x01, 0x01, 0x02, 0x03])
        data, offset = mixer_parameters_structure.decode(message)
        assert data == {
            ATTR_MIXER_PARAMETERS: {
                0: [(0, ParameterValues(value=1, min_value=2, max_value=3))]
            }
        }
        assert offset == 7



"""Contains an ecoSTER class."""

from __future__ import annotations

from pyplumio.const import DeviceType
from pyplumio.devices import PhysicalDevice


class EcoSTER(PhysicalDevice):
    """Represents an ecoSTER thermostat."""

    __slots__ = ()

    address = DeviceType.ECOSTER


__all__ = ["EcoSTER"]



"""Contains test for ecoMAX 860D3-HB custom parameters."""

from __future__ import annotations

from typing import Final

import pytest

from pyplumio.const import STATE_OFF, STATE_ON, ProductType, UnitOfMeasurement
from pyplumio.devices.ecomax import EcoMAX
from pyplumio.frames.responses import EcomaxParametersResponse
from pyplumio.parameters import NumericType
from pyplumio.parameters.ecomax import EcomaxNumber, EcomaxParameter, EcomaxSwitch
from pyplumio.structures.ecomax_parameters import ATTR_ECOMAX_PARAMETERS
from pyplumio.structures.product_info import ATTR_PRODUCT, ProductInfo
from tests.conftest import class_from_json, equal_parameter_value

CELSIUS: Final = UnitOfMeasurement.CELSIUS


@pytest.mark.parametrize(
    ("name", "cls", "value", "min_value", "max_value", "unit_of_measurement"),
    [
        ("__unknown_parameter_1", EcomaxNumber, None, None, None, None),
        ("summer_mode", EcomaxNumber, 1, 0, 2, None),
        ("summer_mode_enable_temp", EcomaxNumber, 16, 5, 30, CELSIUS),
        ("summer_mode_disable_temp", EcomaxNumber, 10, 1, 15, CELSIUS),
        ("water_heater_target_temp", EcomaxNumber, 38, 30, 60, CELSIUS),
        ("min_water_heater_target_temp", EcomaxNumber, 30, 20, 55, CELSIUS),
        ("max_water_heater_target_temp", EcomaxNumber, 60, 25, 80, CELSIUS),
        ("water_heater_feeding_extension", EcomaxNumber, 0, 0, 99, None),
        ("water_heater_work_mode", EcomaxNumber, 2, 0, 2, None),
        ("water_heater_hysteresis", EcomaxNumber, 7, 1, 20, CELSIUS),
        (
            "water_heater_disinfection",
            EcomaxSwitch,
            STATE_OFF,
            STATE_OFF,
            STATE_ON,
            None,
        ),
    ],
)
@class_from_json(
    EcomaxParametersResponse,
    "parameters/ecomax_860d3_hb.json",
    arguments=("message",),
)
async def test_custom_parameters(
    ecomax_860d3_hb: EcomaxParametersResponse,
    ecomax: EcoMAX,
    name: str,
    cls: type[EcomaxParameter],
    value: NumericType | None,
    min_value: NumericType | None,
    max_value: NumericType | None,
    unit_of_measurement: UnitOfMeasurement | None,
) -> None:
    """Test custom parameters for ecoMAX 860D3-HB.

    Thanks @KryspianClash for testdata.
    """
    ecomax.data[ATTR_PRODUCT] = ProductInfo(
        type=ProductType.ECOMAX_P,
        id=48,
        uid="*TEST*",
        logo=48,
        image=2,
        model="ecoMAX 860D3-HB",
    )
    ecomax.handle_frame(ecomax_860d3_hb)
    await ecomax.wait_until_done()
    assert ecomax.data[ATTR_ECOMAX_PARAMETERS] is True

    assert name in ecomax.data
    parameter = ecomax.get_nowait(name, None)
    assert isinstance(parameter, cls)
    if value is None:
        return

    assert equal_parameter_value(parameter.value, value)
    assert equal_parameter_value(parameter.min_value, min_value)
    assert equal_parameter_value(parameter.max_value, max_value)

    if unit_of_measurement is not None:
        assert hasattr(parameter, "unit_of_measurement")
        assert parameter.unit_of_measurement == unit_of_measurement



"""Contains tests for the ecoSTER device."""

import asyncio

from pyplumio.devices.ecoster import EcoSTER
from pyplumio.structures.network_info import NetworkInfo


def test_init() -> None:
    """Test ecoSTER instance."""
    ecoster = EcoSTER(asyncio.Queue(), network=NetworkInfo())
    assert isinstance(ecoster, EcoSTER)



"""Contains an outputs structure decoder."""

from __future__ import annotations

from typing import Any, Final

from pyplumio.data_types import UnsignedInt
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_FAN: Final = "fan"
ATTR_FEEDER: Final = "feeder"
ATTR_HEATING_PUMP: Final = "heating_pump"
ATTR_WATER_HEATER_PUMP: Final = "water_heater_pump"
ATTR_CIRCULATION_PUMP: Final = "circulation_pump"
ATTR_LIGHTER: Final = "lighter"
ATTR_ALARM: Final = "alarm"
ATTR_OUTER_BOILER: Final = "outer_boiler"
ATTR_FAN2_EXHAUST: Final = "fan2_exhaust"
ATTR_FEEDER2: Final = "feeder2"
ATTR_OUTER_FEEDER: Final = "outer_feeder"
ATTR_SOLAR_PUMP: Final = "solar_pump"
ATTR_FIREPLACE_PUMP: Final = "fireplace_pump"
ATTR_GCZ_CONTACT: Final = "gcz_contact"
ATTR_BLOW_FAN1: Final = "blow_fan1"
ATTR_BLOW_FAN2: Final = "blow_fan2"

OUTPUTS: tuple[str, ...] = (
    ATTR_FAN,
    ATTR_FEEDER,
    ATTR_HEATING_PUMP,
    ATTR_WATER_HEATER_PUMP,
    ATTR_CIRCULATION_PUMP,
    ATTR_LIGHTER,
    ATTR_ALARM,
    ATTR_OUTER_BOILER,
    ATTR_FAN2_EXHAUST,
    ATTR_FEEDER2,
    ATTR_OUTER_FEEDER,
    ATTR_SOLAR_PUMP,
    ATTR_FIREPLACE_PUMP,
    ATTR_GCZ_CONTACT,
    ATTR_BLOW_FAN1,
    ATTR_BLOW_FAN2,
)


class OutputsStructure(StructureDecoder):
    """Represents an outputs data structure."""

    __slots__ = ()

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        outputs = UnsignedInt.from_bytes(message, offset)
        return (
            ensure_dict(
                data,
                {
                    output: bool(outputs.value & 2**index)
                    for index, output in enumerate(OUTPUTS)
                },
            ),
            offset + outputs.size,
        )


__all__ = [
    "ATTR_FAN",
    "ATTR_FEEDER",
    "ATTR_HEATING_PUMP",
    "ATTR_WATER_HEATER_PUMP",
    "ATTR_CIRCULATION_PUMP",
    "ATTR_LIGHTER",
    "ATTR_ALARM",
    "ATTR_OUTER_BOILER",
    "ATTR_FAN2_EXHAUST",
    "ATTR_FEEDER2",
    "ATTR_OUTER_FEEDER",
    "ATTR_SOLAR_PUMP",
    "ATTR_FIREPLACE_PUMP",
    "ATTR_GCZ_CONTACT",
    "ATTR_BLOW_FAN1",
    "ATTR_BLOW_FAN2",
    "OutputsStructure",
]
